from __future__ import print_function, division

import os, os.path, re
import logging
import cPickle as pickle


try:
    from configobj import ConfigObj
    import numpy as np
    import matplotlib.pyplot as plt
    from matplotlib import cm
except ImportError:
    ConfigObj, np, plt, cm = (None, None, None, None)

    
from .populations import PopulationSet
from .transitsignal import TransitSignal

from .stars.contrastcurve import ContrastCurveFromFile

from .plotutils import setfig
from .hashutils import hashcombine

try:
    from isochrones import StarModel, BinaryStarModel, TripleStarModel
except ImportError:
    StarModel = None
from .stars.populations import DARTMOUTH

class FPPCalculation(object):
    """
    An object to organize an FPP calculation.

    May be created in one of three ways:

    * Manually building a
      :class:`TransitSignal` and a :class:`PopulationSet`
      and then calling the constructor,
    * Loading from a folder in which the correct data
      files have been saved, using :func:`FPPCalculation.load`, or
    * Reading from a config file, using :func:`FPPCalculation.from_ini`
    

    :param trsig:
        :class:`TransitSignal` object representing the signal
        being modeled.        

    :param popset:
        :class:`PopulationSet` object representing the set
        of models being considered as an explanation for
        the signal.

    :param folder: (optional)
        Folder where likelihood cache, results file, plots, etc.
        are written by default.  
        
    """
    def __init__(self, trsig, popset, folder='.'):
        self.trsig = trsig
        self.name = trsig.name
        self.popset = popset
        self.folder = folder
        if not os.path.exists(self.folder):
            os.makedirs(self.folder)

        lhoodcachefile = os.path.join(self.folder,'lhoodcache.dat')

        self.lhoodcachefile = lhoodcachefile
        for pop in self.popset.poplist:
            pop.lhoodcachefile = lhoodcachefile

    @classmethod
    def from_ini(cls, folder='.',
                 ini_file='fpp.ini', recalc=False,
                 star_ini_file='star.ini', ichrone=DARTMOUTH,
                 **kwargs):
        """
        To enable simple usage, initializes a FPPCalculation from a .ini file

        By default, a file called ``fpp.ini`` will be looked for in the 
        current folder.  Also present must be a ``star.ini`` file that 
        contains the observed properties of the target star.

        ``fpp.ini`` must be of the following form::

            name = k2oi
            ra = 11:30:14.510
            dec = +07:35:18.21

            period = 32.988 #days
            rprs = 0.0534   #Rp/Rstar
            photfile = lc_k2oi.csv

            [constraints]
            maxrad = 10 #exclusion radius [arcsec]
            secthresh = 0.001 #maximum allowed secondary signal depth

            #This variable defines contrast curves
            #ccfiles = Keck_J.cc, Lick_J.cc

        Photfile must be a text file with columns ``(days_from_midtransit,
        flux, flux_err)``.  Both whitespace- and comma-delimited
        will be tried, using ``np.loadtxt``.  Photfile need not be there
        if there is a pickled :class:`TransitSignal` saved in the same
        directory as ``ini_file``, named ``trsig.pkl`` (or another name
        as defined by ``trsig`` keyword in ``.ini`` file). 

        ``star.ini`` should look something like the following::

            B = 15.005, 0.06
            V = 13.496, 0.05
            g = 14.223, 0.05
            r = 12.858, 0.04
            i = 11.661, 0.08
            J = 9.763, 0.03
            H = 9.135, 0.03
            K = 8.899, 0.02
            W1 = 8.769, 0.023
            W2 = 8.668, 0.02
            W3 = 8.552, 0.025
            Kepler = 12.473

            #Teff = 3503, 80
            #feh = 0.09, 0.09
            #logg = 4.89, 0.1            

        Any star properties can be defined; if errors are included
        then they will be used in the :class:`isochrones.StarModel` 
        MCMC fit.
        Spectroscopic parameters (``Teff, feh, logg``) are optional.
        If included, then they will also be included in
        :class:`isochrones.StarModel` fit.  A magnitude for the
        band in which the transit signal is observed (e.g., ``Kepler``)
        is required, though need not have associated uncertainty.


        :param folder: 
            Folder to find configuration files.

        :param ini_file:
            Input configuration file.

        :param star_ini_file:
            Input config file for :class:`isochrones.StarModel` fits.

        :param recalc:
            Whether to re-calculate :class:`PopulationSet`, if a
            ``popset.h5`` file is already present

        :param **kwargs:
            Keyword arguments passed to :class:`PopulationSet`.

        Creates:
        
            * ``trsig.pkl``: the pickled :class:`vespa.TransitSignal` object.
            * ``starfield.h5``: the TRILEGAL field star simulation
            * ``starmodel.h5``: the :class:`isochrones.StarModel` fit
            * ``popset.h5``: the :class:`vespa.PopulationSet` object
              representing the model population simulations.
                    
        """        
        
        if not os.path.isabs(ini_file):
            config = ConfigObj(os.path.join(folder,ini_file))
        else:
            config = ConfigObj(ini_file)

        folder = os.path.abspath(folder)

        #required items
        name = config['name']
        ra, dec = config['ra'], config['dec']
        period = float(config['period'])
        rprs = float(config['rprs'])
        
        #load starmodels if there; 
        # if not, create them.
        if 'starmodel_basename' not in config:
            starmodel_basename = 'dartmouth_starmodel'
        else:
            starmodel_basename = config['starmodel_basename']
        single_starmodel_file = os.path.join(folder,'{}_single.h5'.format(starmodel_basename))
        binary_starmodel_file = os.path.join(folder,'{}_binary.h5'.format(starmodel_basename))
        triple_starmodel_file = os.path.join(folder,'{}_triple.h5'.format(starmodel_basename))

        #Single
        try:
            single_starmodel = StarModel.load_hdf(single_starmodel_file)
            logging.info('Single StarModel loaded from {}'.format(single_starmodel_file))
        except:
            single_starmodel = StarModel.from_ini(ichrone, folder, 
                                           ini_file=star_ini_file)
            logging.info('Fitting single StarModel to {}...'.format(single_starmodel.properties))
            single_starmodel.fit_mcmc()
            single_starmodel.save_hdf(single_starmodel_file)
            triangle_base = os.path.join(folder, '{}_triangle_single'.format(starmodel_basename))
            single_starmodel.triangle_plots(triangle_base)
            logging.info('StarModel fit done.')

        #Binary
        try:
            binary_starmodel = BinaryStarModel.load_hdf(binary_starmodel_file)
            logging.info('BinaryStarModel loaded from {}'.format(binary_starmodel_file))
        except:
            binary_starmodel = BinaryStarModel.from_ini(ichrone, folder,
                                                        ini_file=star_ini_file)
            logging.info('Fitting BinaryStarModel to {}...'.format(binary_starmodel.properties))
            binary_starmodel.fit_mcmc()
            binary_starmodel.save_hdf(binary_starmodel_file)
            triangle_base = os.path.join(folder, '{}_triangle_binary'.format(starmodel_basename))
            binary_starmodel.triangle_plots(triangle_base)
            logging.info('BinaryStarModel fit done.')

        #Triple
        try:
            triple_starmodel = TripleStarModel.load_hdf(triple_starmodel_file)
            logging.info('TripleStarModel loaded from {}'.format(triple_starmodel_file))
        except:
            triple_starmodel = TripleStarModel.from_ini(ichrone, folder,
                                                        ini_file=star_ini_file)
            logging.info('Fitting TripleStarModel to {}...'.format(triple_starmodel.properties))
            triple_starmodel.fit_mcmc()
            triple_starmodel.save_hdf(triple_starmodel_file)
            triangle_base = os.path.join(folder, '{}_triangle_triple'.format(starmodel_basename))
            triple_starmodel.triangle_plots(triangle_base)
            logging.info('TripleStarModel fit done.')

                
        if 'popset' in config:
            popset_file = config['popset']
            if not os.path.isabs(popset_file):
                popset_file = os.path.join(folder, popset_file)
        else:
            popset_file = os.path.join(folder,'popset.h5')

        if 'starfield' in config:
            trilegal_file = config['starfield']
            if not os.path.isabs(trilegal_file):
                trilegal_file = os.path.join(folder, trilegal_file)
        else:
            trilegal_file = os.path.join(folder,'starfield.h5')

        if 'trsig' in config:
            trsig_file = config['trsig']
            if not os.path.isabs(trsig_file):
                trsig_file = os.path.join(folder, trsig_file)
        else:
            trsig_file = os.path.join(folder,'trsig.pkl')
        
            
        #create TransitSignal
        if os.path.exists(trsig_file):
            logging.info('Loading transit signal from {}...'.format(trsig_file))
            trsig = pickle.load(open(trsig_file,'rb'))
        else:
            if 'photfile' not in config:
                raise AttributeError('If transit pickle file (trsig.pkl) '+
                                     'not present, "photfile" must be'+
                                     'defined.')
            if not os.path.isabs(config['photfile']):
                photfile = os.path.join(folder,config['photfile'])
            else:
                photfile = config['photfile']

            logging.info('Reading transit signal photometry ' +
                         'from {}...'.format(photfile))
            try:
                ts, fs, dfs = np.loadtxt(photfile, unpack=True)
            except:
                ts, fs, dfs = np.loadtxt(photfile, delimiter=',', unpack=True)
            
            trsig = TransitSignal(ts, fs, dfs, P=period, name=name)
            logging.info('Fitting transitsignal with MCMC...')
            trsig.MCMC()
            trsig.save(trsig_file)
                        
        #create PopulationSet
        try:
            if recalc:
                raise RuntimeError #just to get to except block
            popset = PopulationSet.load_hdf(popset_file)
            popset['pl'] #should there be a better way to check this? (yes)
            logging.info('PopulationSet loaded from {}'.format(popset_file))
        except:
            if os.path.exists(popset_file):
                logging.warning('{} exists, but regenerating Population Set...'.format(popset_file),
                                exc_info=True)

            popset = PopulationSet(period=period, mags=single_starmodel.mags,
                                   ra=ra, dec=dec,
                                   trilegal_filename=trilegal_file,
                                   starmodel=single_starmodel,
                                   binary_starmodel=binary_starmodel,
                                   triple_starmodel=triple_starmodel,
                                   rprs=rprs,
                                   savefile=popset_file, **kwargs)
            
        
        fpp = cls(trsig, popset, folder=folder)

        maxrad = float(config['constraints']['maxrad'])
        fpp.set_maxrad(maxrad)
        if 'secthresh' in config['constraints']:
            secthresh = float(config['constraints']['secthresh'])
            fpp.apply_secthresh(secthresh)


        #apply contrast curve constraints if present
        if 'ccfiles' in config['constraints']:
            ccfiles = list(config['constraints']['ccfiles'])
            for ccfile in ccfiles:
                if not os.path.isabs(ccfile):
                    ccfile = os.path.join(folder, ccfile)
                m = re.search('(\w+)_(\w+)\.cc',os.path.basename(ccfile))
                if not m:
                    logging.warning('Invalid CC filename ({}); '+
                                     'skipping.'.format(ccfile))
                    continue
                else:
                    band = m.group(2)
                    inst = m.group(1)
                    name = '{} {}-band'.format(inst, band)
                    cc = ContrastCurveFromFile(ccfile, band, name=name)
                    fpp.apply_cc(cc)
        
        return fpp

                   
    def __getattr__(self, attr):
        if attr != 'popset':
            return getattr(self.popset,attr)

    def save(self, overwrite=True):
        """
        Saves PopulationSet and TransitSignal.

        Shouldn't need to use this if you're using
        :func:`FPPCalculation.from_ini`.
        
        Saves :class`PopulationSet` to ``[folder]/popset.h5]``
        and :class:`TransitSignal` to ``[folder]/trsig.pkl``.

        :param overwrite: (optional)
            Whether to overwrite existing files.
        
        """
        self.save_popset(overwrite=overwrite)
        self.save_signal()

    @classmethod
    def load(cls, folder):
        """
        Loads PopulationSet from folder

        ``popset.h5`` and ``trsig.pkl`` must exist in folder.

        :param folder:
            Folder from which to load.
        """
        popset = PopulationSet.load_hdf(os.path.join(folder,'popset.h5'))
        sigfile = os.path.join(folder,'trsig.pkl')
        trsig = pickle.load(open(sigfile, 'rb'))
        return cls(trsig, popset, folder=folder)

    def FPPplots(self, folder=None, format='png', tag=None, **kwargs):
        """
        Make FPP diagnostic plots

        Makes likelihood "fuzz plot" for each model, a FPP summary figure,
        a plot of the :class:`TransitSignal`, and writes a ``results.txt``
        file.

        :param folder: (optional)
            Destination folder for plots/``results.txt``.  Default
            is ``self.folder``.

        :param format: (optional)
            Desired format of figures.  e.g. ``png``, ``pdf``...

        :param tag: (optional)
            If this is provided (string), then filenames will have
            ``_[tag]`` appended to the filename, before the extension.

        :param **kwargs:
            Additional keyword arguments passed to :func:`PopulationSet.lhoodplots`.
            
        """
        if folder is None:
            folder = self.folder

        self.lhoodplots(folder=folder,figformat=format,tag=tag,**kwargs)
        self.FPPsummary(folder=folder,saveplot=True,figformat=format,tag=tag)
        self.plotsignal(folder=folder,saveplot=True,figformat=format)
        self.write_results(folder=folder)

    def plotsignal(self,fig=None,saveplot=True,folder=None,figformat='png',**kwargs):
        """
        Plots TransitSignal

        Calls :func:`TransitSignal.plot`, saves to provided folder.

        :param fig: (optional)
            Argument for :func:`plotutils.setfig`.

        :param saveplot: (optional)
            Whether to save figure.

        :param folder: (optional)
            Folder to which to save plot

        :param figformat: (optional)
            Desired format for figure.

        :param **kwargs:
            Additional keyword arguments passed to :func:`TransitSignal.plot`.
        """
        if folder is None:
            folder = self.folder

        self.trsig.plot(plot_trap=True,fig=fig,**kwargs)
        if saveplot:
            plt.savefig('%s/signal.%s' % (folder,figformat))
            plt.close()

    def write_results(self,folder=None):
        """
        Writes text file of calculation summary.

        :param folder: (optional)
            Folder to which to write ``results.txt``.

            
        """
        if folder is None:
            folder = self.folder
        fout = open(folder+'/'+'results.txt','w')
        for m in self.popset.shortmodelnames:
            fout.write('%s ' % m)
        fout.write('fpV fp FPP\n')
        Ls = {}
        Ltot = 0
        for model in self.popset.modelnames:
            Ls[model] = self.prior(model)*self.lhood(model)
            Ltot += Ls[model]

        line = ''
        for model in self.popset.modelnames:
            line += '%.2e ' % (Ls[model]/Ltot)
        line += '%.3g %.3f %.2e\n' % (self.fpV(),self.priorfactors['fp_specific'],self.FPP())

        fout.write(line)
        fout.close()

    def save_popset(self,filename='popset.h5',**kwargs):
        """Saves the PopulationSet

        Calls :func:`PopulationSet.save_hdf`.
        """
        self.popset.save_hdf(os.path.join(self.folder,filename))

    def save_signal(self,filename=None):
        """
        Saves TransitSignal.

        Calls :func:`TransitSignal.save`; default filename is
        ``trsig.pkl`` in ``self.folder``.
        """
        if filename is None:
            filename = os.path.join(self.folder,'trsig.pkl')
        self.trsig.save(filename)

    def FPPsummary(self,fig=None,figsize=(10,8),saveplot=False,folder='.',
                   starinfo=True,siginfo=True,
                   priorinfo=True,constraintinfo=True,
                   tag=None,simple=False,figformat='png'):
        """
        Makes FPP summary plot

        .. note::

            This is due for updates/improvements.
        
        :param fig, figsize: (optional)
            Arguments for :func:`plotutils.setfig`.

        :param saveplot: (optional)
            Whether to save figure.  Default is ``False``.
        
        :param folder: (optional)
            Folder to which to save plot; default is current working dir.

        :param figformat: (optional)
            Desired format of saved figure.
            
        """
        if simple:
            starinfo = False
            siginfo = False
            priorinfo = False
            constraintinfo = False

        setfig(fig,figsize=figsize)
        # three pie charts
        priors = []; lhoods = []; Ls = []
        for model in self.popset.modelnames:
            priors.append(self.prior(model))
            lhoods.append(self.lhood(model))
            if np.isnan(priors[-1]):
                raise ValueError('{} prior is nan; priorfactors={}'.format(model,
                                                                           self.priorfactors))
            Ls.append(priors[-1]*lhoods[-1])
        priors = np.array(priors)
        lhoods = np.array(lhoods)

        Ls = np.array(Ls)
        logging.debug('modelnames={}'.format(self.popset.modelnames))
        logging.debug('priors={}'.format(priors))
        logging.debug('lhoods={}'.format(lhoods))

        #colors = ['b','g','r','m','c']
        nmodels = len(self.popset.modelnames)
        colors = [cm.jet(1.*i/nmodels) for i in range(nmodels)]
        legendprop = {'size':11}

        ax1 = plt.axes([0.15,0.45,0.35,0.43])
        try:
            plt.pie(priors/priors.sum(),colors=colors)
            labels = []
            for i,model in enumerate(self.popset.modelnames):
                labels.append('%s: %.1e' % (model,priors[i]))
            plt.legend(labels,bbox_to_anchor=(-0.25,-0.1),loc='lower left',prop=legendprop)
            plt.title('Priors')
        except:
            msg = 'Error calculating priors.\n'
            for i,mod in enumerate(self.popset.shortmodelnames):
                msg += '%s: %.1e' % (model,priors[i])
            plt.annotate(msg, xy=(0.5,0.5), xycoords='axes fraction')
            

        ax2 = plt.axes([0.5,0.45,0.35,0.43])
        try:
            plt.pie(lhoods/lhoods.sum(),colors=colors)            
            labels = []
            for i,model in enumerate(self.popset.modelnames):
                labels.append('%s: %.1e' % (model,lhoods[i]))
            plt.legend(labels,bbox_to_anchor=(1.25,-0.1),loc='lower right',prop=legendprop)
            plt.title('Likelihoods')
        except:
            msg = 'Error calculating lhoods.\n'
            for i,mod in enumerate(self.popset.shortmodelnames):
                msg += '%s: %.1e' % (model,lhoods[i])
            plt.annotate(msg, xy=(0.5,0.5), xycoords='axes fraction')

        ax3 = plt.axes([0.3,0.03,0.4,0.5])
        try:
            plt.pie(Ls/Ls.sum(),colors=colors)
            labels = []
            #for i,model in enumerate(['eb','heb','bgeb','bgpl','pl']):
            for i,model in enumerate(self.popset.modelnames):
                labels.append('%s: %.3f' % (model,Ls[i]/Ls.sum()))
            plt.legend(labels,bbox_to_anchor=(1.6,0.44),loc='right',prop={'size':14},shadow=True)
            plt.annotate('Final Probability',xy=(0.5,-0.01),ha='center',xycoords='axes fraction',fontsize=18)
        except:
            msg = 'Error calculating final probabilities.\n'
            plt.annotate(msg, xy=(0.5,0.5), xycoords='axes fraction')
            

        """
        #starpars = 'Star parameters used\nin simulations'
        starpars = ''
        if 'M' in self['heb'].stars.keywords and 'DM_P' in self.keywords:
            starpars += '\n$M/M_\odot = %.2f^{+%.2f}_{-%.2f}$' % (self['M'],self['DM_P'],self['DM_N'])
        else:
            starpars += '\n$(M/M_\odot = %.2f \pm %.2f)$' % (self['M'],0)  #this might not always be right?

        if 'DR_P' in self.keywords:
            starpars += '\n$R/R_\odot = %.2f^{+%.2f}_{-%.2f}$' % (self['R'],self['DR_P'],self['DR_N'])
        else:
            starpars += '\n$R/R_\odot = %.2f \pm %.2f$' % (self['R'],self['DR'])

        if 'FEH' in self.keywords:
            if 'DFEH_P' in self.keywords:
                starpars += '\n$[Fe/H] = %.2f^{+%.2f}_{-%.2f}$' % (self['FEH'],self['DFEH_P'],self['DFEH_N'])
            else:
                starpars += '\n$[Fe/H] = %.2f \pm %.2f$' % (self['FEH'],self['DFEH'])
        for kw in self.keywords:
            if re.search('-',kw):
                try:
                    starpars += '\n$%s = %.2f (%.2f)$ ' % (kw,self[kw],self['COLORTOL'])
                except TypeError:
                    starpars += '\n$%s = %s (%.2f)$ ' % (kw,self[kw],self['COLORTOL'])
                    
        #if 'J-K' in self.keywords:
        #    starpars += '\n$J-K = %.2f (%.2f)$ ' % (self['J-K'],self['COLORTOL'])
        #if 'G-R' in self.keywords:
        #    starpars += '\n$g-r = %.2f (%.2f)$' % (self['G-R'],self['COLORTOL'])
        if starinfo:
            plt.annotate(starpars,xy=(0.03,0.91),xycoords='figure fraction',va='top')

        #p.annotate('Star',xy=(0.04,0.92),xycoords='figure fraction',va='top')

        priorpars = r'$f_{b,short} = %.2f$  $f_{trip} = %.2f$' % (self.priorfactors['fB']*self.priorfactors['f_Pshort'],
                                                                self.priorfactors['ftrip'])
        if 'ALPHA' in self.priorfactors:
            priorpars += '\n'+r'$f_{pl,bg} = %.2f$  $\alpha_{pl,bg} = %.1f$' % (self.priorfactors['fp'],self['ALPHA'])
        else:
            priorpars += '\n'+r'$f_{pl,bg} = %.2f$  $\alpha_1,\alpha_2,r_b = %.1f,%.1f,%.1f$' % \
                         (self.priorfactors['fp'],self['bgpl'].stars.keywords['ALPHA1'],
                          self['bgpl'].stars.keywords['ALPHA2'],
                          self['bgpl'].stars.keywords['RBREAK'])
            
        rbin1,rbin2 = self['RBINCEN']-self['RBINWID'],self['RBINCEN']+self['RBINWID']
        priorpars += '\n$f_{pl,specific} = %.2f, \in [%.2f,%.2f] R_\oplus$' % (self.priorfactors['fp_specific'],rbin1,rbin2)
        priorpars += '\n$r_{confusion} = %.1f$"' % sqrt(self.priorfactors['area']/pi)
        if self.priorfactors['multboost'] != 1:
            priorpars += '\nmultiplicity boost = %ix' % self.priorfactors['multboost']
        if priorinfo:
            plt.annotate(priorpars,xy=(0.03,0.4),xycoords='figure fraction',va='top')
        

        sigpars = ''
        sigpars += '\n$P = %s$ d' % self['P']
        depth,ddepth = self.trsig.depthfit
        sigpars += '\n$\delta = %i^{+%i}_{-%i}$ ppm' % (depth*1e6,ddepth[1]*1e6,ddepth[0]*1e6)
        dur,ddur = self.trsig.durfit
        sigpars += '\n$T = %.2f^{+%.2f}_{-%.2f}$ h' % (dur*24.,ddur[1]*24,ddur[0]*24)
        slope,dslope = self.trsig.slopefit
        sigpars += '\n'+r'$T/\tau = %.1f^{+%.1f}_{-%.1f}$' % (slope,dslope[1],dslope[0])
        sigpars += '\n'+r'$(T/\tau)_{max} = %.1f$' % (self.trsig.maxslope)
        if siginfo:
            plt.annotate(sigpars,xy=(0.81,0.91),xycoords='figure fraction',va='top')
        
            #p.annotate('${}^a$Not used for FP population simulations',xy=(0.02,0.02),
            #           xycoords='figure fraction',fontsize=9)
        """

        constraints = 'Constraints:'
        for c in self.popset.constraints:
            try:
                constraints += '\n  %s' % self['heb'].constraints[c]
            except KeyError:
                constraints += '\n  %s' % self['beb'].constraints[c]                
        if constraintinfo:
            plt.annotate(constraints,xy=(0.03,0.22),xycoords='figure fraction',
                         va='top',color='red')
            
        odds = 1./self.FPP()

        if odds > 1e6:
            fppstr = 'FPP: < 1 in 1e6' 
        elif np.isfinite(odds):
            fppstr = 'FPP: 1 in %i' % odds
        else:
            fppstr = 'FPP calculation failed.'

        plt.annotate('$f_{pl,V} = %.3f$\n%s' % (self.fpV(),fppstr),xy=(0.7,0.02),
                     xycoords='figure fraction',fontsize=16,va='bottom')

        plt.suptitle(self.trsig.name,fontsize=22)

        #if not simple:
        #    plt.annotate('n = %.0e' % self.n,xy=(0.5,0.85),xycoords='figure fraction',
        #                 fontsize=14,ha='center')

        if saveplot:
            if tag is not None:
                plt.savefig('%s/FPPsummary_%s.%s' % (folder,tag,figformat))
            else:
                plt.savefig('%s/FPPsummary.%s' % (folder,figformat))
            plt.close()


    def lhoodplots(self,folder='.',tag=None,figformat='png',**kwargs):
        """
        Make a plot of the likelihood for each model in PopulationSet

        """
        Ltot = 0

        for model in self.popset.modelnames:
            Ltot += self.prior(model)*self.lhood(model)
        
        for model in self.popset.shortmodelnames:
            self.lhoodplot(model,Ltot=Ltot,**kwargs)
            if tag is None:
                plt.savefig('%s/%s.%s' % (folder,model,figformat))
            else:
                plt.savefig('%s/%s_%s.%s' % (folder,model,figformat))
            plt.close()

    def lhoodplot(self,model,suptitle='',**kwargs):
        """
        Make a plot of the likelihood for a given model.
        """
        if suptitle=='':
            suptitle = self[model].model
        self[model].lhoodplot(self.trsig,colordict=self.popset.colordict,
                              suptitle=suptitle,**kwargs)

    def calc_lhoods(self,**kwargs):
        logging.debug('Calculating likelihoods...')
        for pop in self.popset.poplist:
            L = pop.lhood(self.trsig,**kwargs)
            logging.debug('%s: %.2e' % (pop.model,L))

    def __hash__(self):
        return hashcombine(self.popset, self.trsig)

    def __getitem__(self,model):
        return self.popset[model]

    def prior(self,model):
        """
        Return the prior for a given model.
        """
        return self[model].prior

    def lhood(self,model,**kwargs):
        """
        Return the likelihood for a given model.
        """
        return self[model].lhood(self.trsig,
                                 **kwargs)

    def Pval(self,skipmodels=None):
        Lfpp = 0
        if skipmodels is None:
            skipmodels = []
        logging.debug('evaluating likelihoods for %s' % self.trsig.name)
        
        for model in self.popset.modelnames:
            if model=='Planets':
                continue
            if model not in skipmodels:
                prior = self.prior(model)
                lhood = self.lhood(model)
                Lfpp += prior*lhood
                logging.debug('%s: %.2e = %.2e (prior) x %.2e (lhood)' % (model,prior*lhood,prior,lhood))
        prior = self.prior('pl')
        lhood = self.lhood('pl')
        Lpl = prior*lhood
        logging.debug('planet: %.2e = %.2e (prior) x %.2e (lhood)' % (prior*lhood,prior,lhood))
        return Lpl/Lfpp/self['pl'].priorfactors['fp_specific']

    def fpV(self,FPPV=0.005,skipmodels=None):
        P = self.Pval(skipmodels=skipmodels)
        return (1-FPPV)/(P*FPPV)

    def FPP(self,skipmodels=None):
        """
        Return the false positive probability (FPP)
        """
        Lfpp = 0
        if skipmodels is None:
            skipmodels = []
        logging.debug('evaluating likelihoods for %s' % self.trsig.name)
        for model in self.popset.modelnames:
            if model=='Planets':
                continue
            if model not in skipmodels:
                prior = self.prior(model)
                lhood = self.lhood(model)
                Lfpp += prior*lhood
                logging.debug('%s: %.2e = %.2e (prior) x %.2e (lhood)' % (model,prior*lhood,prior,lhood))
        prior = self.prior('pl')
        lhood = self.lhood('pl')
        Lpl = prior*lhood
        logging.debug('planet: %.2e = %.2e (prior) x %.2e (lhood)' % (prior*lhood,prior,lhood))
        return 1 - Lpl/(Lpl + Lfpp)




from __future__ import division, print_function

import os,os.path
import logging
import pickle

try:
    import numpy as np
    import pandas as pd
    import matplotlib.pyplot as plt
    import numpy.random as rand
    from scipy.stats import gaussian_kde
except ImportError:
    np, pd, plt, rand = (None, None, None, None)
    gaussian_kde = None
    
try:
    import acor
except ImportError:
    logging.warning('acor not available')
    
from .plotutils import setfig
from .hashutils import hashcombine, hasharray

from .transit_basic import traptransit, fit_traptransit, traptransit_MCMC
from .statutils import kdeconf, qstd, conf_interval

def load_pkl(filename):
    fin = open(filename, 'rb')
    return pickle.load(fin)

class TransitSignal(object):
    """A phased-folded transit signal.

    Epoch of the transit at 0, 'continuum' set at 1.
    
    :param ts, fs, dfs:
        Times (days from mid-transit), fluxes (relative to 1),
        flux uncertainties.  dfs optional

    :param P:
        Orbital period.

    :param p0: (optional)
        Initial guess for least-squares trapezoid fit.
        If not provided, then some decent guess will be made
        (which is better on made-up data than real...)

    :param name: (optional)
        Name of the signal.

    :param maxslope: (optional)
        Upper limit to use for "slope" parameter (T/tau)
        in the MCMC fitting of signal.  Default is 30.


    .. note:: The implementation of this object can use some refactoring;
         as it is directly translated from some older code.  As
         such, not all methods/attributes are well documented.

         
    """
    def __init__(self,ts,fs,dfs=None,P=None,p0=None,name='',maxslope=30):

        ts = np.atleast_1d(ts)
        fs = np.atleast_1d(fs)

        inds = ts.argsort()
        self.ts = ts[inds]
        self.fs = fs[inds]
        self.name = name
        self.P = P

        self.maxslope = maxslope
        if type(P) == type(np.array([1])):
            self.P = P[0]

        #set default best-guess trapezoid parameters
        if p0 is None:
            depth = 1 - fs.min()
            duration = (fs < (1-0.01*depth)).sum()/float(len(fs)) * (ts[-1] - ts[0])
            tc0 = ts[fs.argmin()]
            p0 = np.array([duration,depth,5.,tc0])

        tfit = fit_traptransit(ts,fs,p0)

        if dfs is None:
            dfs = (self.fs - traptransit(self.ts,tfit)).std()
        if np.size(dfs)==1:
            dfs = np.ones(len(self.ts))*dfs
        self.dfs = dfs

        self.dur,self.depth,self.slope,self.center = tfit
        self.trapfit = tfit

        logging.debug('trapezoidal leastsq fit: {}'.format(self.trapfit))

        self.hasMCMC=False

    def save_hdf(self, filename, path=''):
        """
        Save transitsignal info using HDF...not yet implemented.

        .. note::

           Refactoring plan is to re-write saving to use HDF
           instead of pickle.
        """
        raise NotImplementedError
        
    
    def save_pkl(self, filename):
        """
        Pickles TransitSignal.
        """
        fout = open(filename, 'wb')
        pickle.dump(self, fout)
        fout.close()

    #eventually make this save_hdf
    def save(self, filename):
        """
        Calls save_pkl function.
        """
        self.save_pkl(filename)
        
    def __eq__(self,other):
        return hash(self) == hash(other)

    def __hash__(self):
        key =  hashcombine(hasharray(self.ts),
                           hasharray(self.fs),
                           self.P,
                           self.maxslope)
        if self.hasMCMC:
            key = hashcombine(key, hasharray(self.slopes),
                              hasharray(self.durs),
                              hasharray(self.logdeps))
        return key
        
    def plot(self, fig=None, plot_trap=False, name=False, trap_color='g',
             trap_kwargs=None, **kwargs):
        """
        Makes a simple plot of signal

        :param fig: (optional)
            Argument for :func:`plotutils.setfig`.

        :param plot_trap: (optional)
            Whether to plot the (best-fit least-sq) trapezoid fit.

        :param name: (optional)
            Whether to annotate plot with the name of the signal;
            can be ``True`` (in which case ``self.name`` will be
            used), or any arbitrary string.

        :param trap_color: (optional)
            Color of trapezoid fit line.

        :param trap_kwargs: (optional)
            Keyword arguments to pass to trapezoid fit line.

        :param **kwargs: (optional)
            Additional keyword arguments passed to ``plt.plot``.

            
        """

        setfig(fig)
        
        plt.plot(self.ts,self.fs,'.',**kwargs)

        if plot_trap and hasattr(self,'trapfit'):
            if trap_kwargs is None:
                trap_kwargs = {}
            plt.plot(self.ts, traptransit(self.ts,self.trapfit),
                     color=trap_color, **trap_kwargs)

        if name is not None:
            if type(name)==type(''):
                text = name
            else:
                text = self.name
            plt.annotate(text,xy=(0.1,0.1),xycoords='axes fraction',fontsize=22)

        if hasattr(self,'depthfit') and not np.isnan(self.depthfit[0]):
            lo = 1 - 3*self.depthfit[0]
            hi = 1 + 2*self.depthfit[0]
        else:
            lo = 1
            hi = 1

        sig = qstd(self.fs,0.005)
        hi = max(hi,self.fs.mean() + 7*sig)
        lo = min(lo,self.fs.mean() - 7*sig)
        logging.debug('lo={}, hi={}'.format(lo,hi))
        plt.ylim((lo,hi))
        plt.xlabel('time [days]')
        plt.ylabel('Relative flux')

    def MCMC(self, niter=500, nburn=200, nwalkers=200, threads=1,
             fit_partial=False, width=3, savedir=None, refit=False,
             thin=10, conf=0.95, maxslope=30, debug=False, p0=None):
        """
        Fit transit signal to trapezoid model using MCMC

        .. note:: As currently implemented, this method creates a
            bunch of attributes relevant to the MCMC fit; I plan
            to refactor this to define those attributes as properties
            so as not to have their creation hidden away here.  I plan
            to refactor how this works.
        """
        if fit_partial:
            wok = np.where((np.absolute(self.ts-self.center) < (width*self.dur)) &
                           ~np.isnan(self.fs))
        else:
            wok = np.where(~np.isnan(self.fs))

        if savedir is not None:
            if not os.path.exists(savedir):
                os.mkdir(savedir)

        alreadydone = True
        alreadydone &= savedir is not None
        alreadydone &= os.path.exists('%s/ts.npy' % savedir)
        alreadydone &= os.path.exists('%s/fs.npy' % savedir)

        if savedir is not None and alreadydone:
            ts_done = np.load('%s/ts.npy' % savedir)
            fs_done = np.load('%s/fs.npy' % savedir)
            alreadydone &= np.all(ts_done == self.ts[wok])
            alreadydone &= np.all(fs_done == self.fs[wok])
        
        if alreadydone and not refit:
            logging.info('MCMC fit already done for %s.  Loading chains.' % self.name)
            Ts = np.load('%s/duration_chain.npy' % savedir)
            ds = np.load('%s/depth_chain.npy' % savedir)
            slopes = np.load('%s/slope_chain.npy' % savedir)
            tcs = np.load('%s/tc_chain.npy' % savedir)
        else:
            logging.info('Fitting data to trapezoid shape with MCMC for %s....' % self.name)
            if p0 is None:
                p0 = self.trapfit.copy()
                p0[0] = np.absolute(p0[0])
                if p0[2] < 2:
                    p0[2] = 2.01
                if p0[1] < 0:
                    p0[1] = 1e-5
            logging.debug('p0 for MCMC = {}'.format(p0))
            sampler = traptransit_MCMC(self.ts[wok],self.fs[wok],self.dfs[wok],
                                        niter=niter,nburn=nburn,nwalkers=nwalkers,
                                        threads=threads,p0=p0,return_sampler=True,
                                        maxslope=maxslope)

            Ts,ds,slopes,tcs = (sampler.flatchain[:,0],sampler.flatchain[:,1],
                                sampler.flatchain[:,2],sampler.flatchain[:,3])

            self.sampler = sampler
            if savedir is not None:
                np.save('%s/duration_chain.npy' % savedir,Ts)
                np.save('%s/depth_chain.npy' % savedir,ds)
                np.save('%s/slope_chain.npy' % savedir,slopes)
                np.save('%s/tc_chain.npy' % savedir,tcs)
                np.save('%s/ts.npy' % savedir,self.ts[wok])
                np.save('%s/fs.npy' % savedir,self.fs[wok])

        if debug:
            print(Ts)
            print(ds)
            print(slopes)
            print(tcs)

        N = len(Ts)
        self.Ts_acor = acor.acor(Ts)[1]
        self.ds_acor = acor.acor(ds)[1]
        self.slopes_acor = acor.acor(slopes)[1]
        self.tcs_acor = acor.acor(tcs)[1]
        self.fit_converged = True
        for t in [self.Ts_acor,self.ds_acor,
                  self.slopes_acor,self.tcs_acor]:
            if t > 0.1*N:
                self.fit_converged = False


        ok = (Ts > 0) & (ds > 0) & (slopes > 0) & (slopes < self.maxslope)
        logging.debug('trapezoidal fit has {} good sample points'.format(ok.sum()))
        if ok.sum()==0:
            if (Ts > 0).sum()==0:
                #logging.debug('{} points with Ts > 0'.format((Ts > 0).sum()))
                logging.debug('{}'.format(Ts))
                raise MCMCError('{}: 0 points with Ts > 0'.format(self.name))
            if (ds > 0).sum()==0:
                #logging.debug('{} points with ds > 0'.format((ds > 0).sum()))
                logging.debug('{}'.format(ds))
                raise MCMCError('{}: 0 points with ds > 0'.format(self.name))
            if (slopes > 0).sum()==0:
                #logging.debug('{} points with slopes > 0'.format((slopes > 0).sum()))
                logging.debug('{}'.format(slopes))
                raise MCMCError('{}: 0 points with slopes > 0'.format(self.name))
            if (slopes < self.maxslope).sum()==0:
                #logging.debug('{} points with slopes < maxslope ({})'.format((slopes < self.maxslope).sum(),self.maxslope))
                logging.debug('{}'.format(slopes))
                raise MCMCError('{} points with slopes < maxslope ({})'.format((slopes < self.maxslope).sum(),self.maxslope))


        durs,deps,logdeps,slopes = (Ts[ok],ds[ok],np.log10(ds[ok]),
                                              slopes[ok])
        
        
        inds = (np.arange(len(durs)/thin)*thin).astype(int)
        durs,deps,logdeps,slopes = (durs[inds],deps[inds],logdeps[inds],
                                              slopes[inds])

        self.durs,self.deps,self.logdeps,self.slopes = (durs,deps,logdeps,slopes)

        self._make_kde(conf=conf)

        self.hasMCMC = True

    def _make_kde(self, conf=0.95):

        self.durkde = gaussian_kde(self.durs)
        self.depthkde = gaussian_kde(self.deps)
        self.slopekde = gaussian_kde(self.slopes)
        self.logdepthkde = gaussian_kde(self.logdeps)


        if self.fit_converged:
            try:
                durconf = kdeconf(self.durkde,conf)
                depconf = kdeconf(self.depthkde,conf)
                logdepconf = kdeconf(self.logdepthkde,conf)
                slopeconf = kdeconf(self.slopekde,conf)
            except:
                raise
                raise MCMCError('Error generating confidence intervals...fit must not have worked.')

            durmed = np.median(self.durs)
            depmed = np.median(self.deps)
            logdepmed = np.median(self.logdeps)
            slopemed = np.median(self.slopes)

            self.durfit = (durmed,np.array([durmed-durconf[0],durconf[1]-durmed]))
            self.depthfit = (depmed,np.array([depmed-depconf[0],depconf[1]-depmed]))
            self.logdepthfit = (logdepmed,np.array([logdepmed-logdepconf[0],logdepconf[1]-logdepmed]))
            self.slopefit = (slopemed,np.array([slopemed-slopeconf[0],slopeconf[1]-slopemed]))

        else:
            self.durfit = (np.nan,np.nan,np.nan)
            self.depthfit = (np.nan,np.nan,np.nan)
            self.logdepthfit = (np.nan,np.nan,np.nan)
            self.slopefit = (np.nan,np.nan,np.nan)


        points = np.array([self.durs,self.logdeps,self.slopes])
        self.kde = gaussian_kde(points)
        

class TransitSignal_FromSamples(TransitSignal):
    """Use this if all you have is the trapezoid-fit samples
    """
    def __init__(self, period, durs, depths, slopes, 
                 name='', **kwargs):
        self.period = period
        self.durs = durs
        self.deps = depths
        self.logdeps = np.log10(depths)
        self.slopes = slopes
        self.hasMCMC = True
        self.fit_converged = True #better be
        self._make_kde()
        self.name = name

    def MCMC(self, *args, **kwargs):
        pass

    def plot(self, *args, **kwargs):
        pass

    def __hash__(self):
        return hashcombine(self.period, hasharray(self.durs),
                           hasharray(self.deps),
                           hasharray(self.slopes))

class TransitSignal_DF(TransitSignal):
    def __init__(self, df, columns=['t','f','e_f'], **kwargs):
        t_col, f_col, e_f_col = columns
        t = df[t_col]
        f = df[f_col]
        if e_f_col in df:
            e_f = df[e_f_col]
        else:
            e_f = None
        TransitSignal.__init__(self, t, f, e_f, **kwargs)

class TransitSignal_ASCII(TransitSignal):
    def __init__(self, filename, cols=(0,1), err_col=2, **kwargs):
        t, f = np.loadtxt(filename, usecols=cols, unpack=True)
        try:
            e_f = np.loadtxt(filename, usecols=(err_col,))
        except:
            e_f = None
        TransitSignal.__init__(self, t, f, e_f, **kwargs)
        
        
############# Exceptions ##############3

class MCMCError(Exception):
    pass



from __future__ import division,print_function

import sys,re,os
import logging

try:
    import numpy as np
    from astropy.coordinates import SkyCoord,Angle
    import numpy.random as rand
    import matplotlib.pyplot as plt

    import pandas as pd

    from astropy import units as u
    from astropy.units.quantity import Quantity
    from astropy import constants as const
    MSUN = const.M_sun.cgs.value
    AU = const.au.cgs.value
    DAY = 86400
    G = const.G.cgs.value
    
except ImportError:
    np = None
    SkyCoord, Angle = (None, None)
    rand = None
    plt = None
    pd = None
    u = None
    Quantity = None
    const = None
    MSUN, AU, DAY, G = (None, None, None, None)

    
from .utils import semimajor,random_spherepos,orbitproject,orbit_posvel

from ..plotutils import setfig
from ..hashutils import hashcombine, hashdf


class TripleOrbitPopulation(object):             
    """
    Stars 2 and 3 orbit each other (short orbit), far from star 1 (long orbit)

    This object defines the orbits of a triple star system,
    with orbits calculated assuming the "long" orbit does not perturb
    the "short" orbit, which will not be true in the long run, but should
    be true over short timescales as long as ``Plong >> Pshort``. 

    A :class:`TripleOrbitPopulation` is essentially made up of two
    :class:`OrbitPopulation` objects: one for the "long" orbit
    and one for the "short."

    :param M1,M2,M3:
        Masses of stars.  Stars 2 and 3 are in a short orbit, far away from star 1.
        If not :class:`astropy.units.Quantity` objects, then assumed to be
        in solar mass units.  May be single value or array-like.

    :param Plong,Pshort: 
        Orbital Periods.  Plong is orbital period of 2+3 and 1; Pshort is orbital
        period of 2 and 3.  If not :class:`astropy.units.Quantity` objects,
        assumed to be in days.  Can be single value or array-like.
        N.B. If any item in Pshort happens to be
        longer than the corresponding item in Plong, they will be switched. 

    :param ecclong,eccshort: (optional) 
        Eccentricities.  Same story (long vs. short).  Default=0 (circular).
        Can be single value or array-like.

    :param n: (optional)
        Number of systems to simulate (if ``M1``, ``M2``, ``M3`` aren't
        arrays of size > 1 already).

    :param mean_anomaly_short,mean_anomaly_long: (optional)
        Mean anomalies.  This is only passed if you need to restore a
        particular specific configuration (i.e., a particular saved simulation),
        e.g., as done by :func:`TripleOrbitPopulation.from_df`.
        If not provided, then randomized on (0, 2pi).

    :param obsx_short,obsy_short,obsz_short: (optional)
        "Observer" positions for the short orbit.  Also only passed for purposes
        of restoring configuration.

    :param obsx_long,obsy_long,obsz_long: (optional)
        "Observer" positions for long orbit.  Also only passed for purposes of
        restoring configuration.

    :param obspos_short,obspos_long: (optional)
        "Observer positions for short and long orbits, provided
        as :class:`astropy.SkyCoord` objects.  These will replace
        obsx_short/long, obsy_short/long, obsz_short/long parameters if present.

    """
    def __init__(self,M1,M2,M3,Plong,Pshort,ecclong=0,eccshort=0,n=None,
                 mean_anomaly_long=None,obsx_long=None,obsy_long=None,obsz_long=None,
                 obspos_long=None,
                 mean_anomaly_short=None,obsx_short=None,
                 obsy_short=None,obsz_short=None,
                 obspos_short=None):

        Pshort, Plong = (np.minimum(Pshort,Plong), np.maximum(Pshort,Plong))
        #if Plong < Pshort:
        #    Pshort,Plong = (Plong, Pshort)

        self.orbpop_long = OrbitPopulation(M1,M2+M3,Plong,ecc=ecclong,n=n,
                                           mean_anomaly=mean_anomaly_long,
                                           obsx=obsx_long,obsy=obsy_long,obsz=obsz_long)

        self.orbpop_short = OrbitPopulation(M2,M3,Pshort,ecc=eccshort,n=n,
                                           mean_anomaly=mean_anomaly_short,
                                           obsx=obsx_short,obsy=obsy_short,obsz=obsz_short)

    def __hash__(self):
        return hashcombine(self.orbpop_long, self.orbpop_short)
        
    @property
    def RV(self):
        """
        Instantaneous RV of star 1 with respect to system center-of-mass
        """
        return self.RV_1
        
    @property
    def RV_1(self):
        """
        Instantaneous RV of star 1 with respect to system center-of-mass
        """
        return self.orbpop_long.RV * (self.orbpop_long.M2 / (self.orbpop_long.M1 + self.orbpop_long.M2))

    @property
    def RV_2(self):
        """
        Instantaneous RV of star 2 with respect to system center-of-mass
        """
        return -self.orbpop_long.RV * (self.orbpop_long.M1 /
                                        (self.orbpop_long.M1 + self.orbpop_long.M2)) +\
                self.orbpop_short.RV_com1
                
    @property
    def RV_3(self):
        """
        Instantaneous RV of star 3 with respect to system center-of-mass
        """
        return -self.orbpop_long.RV * (self.orbpop_long.M1 / (self.orbpop_long.M1 + self.orbpop_long.M2)) +\
            self.orbpop_short.RV_com2

    @property
    def Rsky(self):
        """
        Projected separation of star 2+3 pair from star 1  [projected AU]
        """
        return self.orbpop_long.Rsky

    def dRV(self,dt):
        """
        Returns difference in RVs (separated by time dt) of star 1.

        :param dt:
            Time separation for which to compute RV change.  If not an
            :class:`astropy.units.Quantity` object, then assumed to be in days.

        """
        return self.dRV_1(dt)
    
    def dRV_1(self,dt):
        """
        Returns difference in RVs (separated by time dt) of star 1.

        :param dt:
            Time separation for which to compute RV change.  If not an
            :class:`astropy.units.Quantity` object, then assumed to be in days.

        """
        return self.orbpop_long.dRV(dt,com=True)

    def dRV_2(self,dt):
        """
        Returns difference in RVs (separated by time dt) of star 2.

        :param dt:
            Time separation for which to compute RV change.  If not an
            :class:`astropy.units.Quantity` object, then assumed to be in days.

        """
        return -self.orbpop_long.dRV(dt) * (self.orbpop_long.M1/(self.orbpop_long.M1 + self.orbpop_long.M2)) +\
            self.orbpop_short.dRV(dt,com=True)

    def dRV_3(self,dt):
        """
        Returns difference in RVs (separated by time dt) of star 3.

        :param dt:
            Time separation for which to compute RV change.  If not an
            :class:`astropy.units.Quantity` object, then assumed to be in days.

        """
        return -self.orbpop_long.dRV(dt) * (self.orbpop_long.M1/(self.orbpop_long.M1 + self.orbpop_long.M2)) -\
            self.orbpop_short.dRV(dt) * (self.orbpop_short.M1/(self.orbpop_short.M1 + self.orbpop_short.M2))

    def save_hdf(self,filename,path=''):
        """
        Save to HDF5 file in desired path.
        """
        self.orbpop_long.save_hdf(filename,'{}/long'.format(path))
        self.orbpop_short.save_hdf(filename,'{}/short'.format(path))
        

    def __add__(self, other):
        if type(self) != type(other):
            raise TypeError('Can only add like types of TripleOrbitPopulation')

        newdf_long = pd.concat((self.orbpop_long.dataframe, other.orbpop_long.dataframe))
        newdf_short = pd.concat((self.orbpop_short.dataframe, other.orbpop_short.dataframe))

        return TripleOrbitPopulation_FromDF(newdf_long, newdf_short)


    @classmethod
    def from_df(cls, df_long, df_short):
        """
        Builds TripleOrbitPopulation from DataFrame

        ``DataFrame`` objects must be of appropriate form to pass
        to :func:`OrbitPopulation.from_df`.

        :param df_long, df_short:
            :class:`pandas.DataFrame` objects to pass to
            :func:`OrbitPopulation.from_df`.
            
        """
        pop = cls(1,1,1,1,1) #dummy population
        pop.orbpop_long = OrbitPopulation.from_df(df_long)
        pop.orbpop_short = OrbitPopulation.from_df(df_short)
        return pop
        
    @classmethod
    def load_hdf(cls, filename, path=''):
        """
        Load TripleOrbitPopulation from saved .h5 file.

        :param filename:
            HDF file name.

        :param path:
            Path within HDF file where data is stored.
            
        """
        df_long = pd.read_hdf(filename,'{}/long/df'.format(path), autoclose=True)
        df_short = pd.read_hdf(filename,'{}/short/df'.format(path), autoclose=True)
        return cls.from_df(df_long, df_short)
        
            
class OrbitPopulation(object):
    """Population of orbits.

    :param M1,M2: 
        Primary and secondary masses (if not ``Quantity``, 
        assumed to be in solar masses).  Can be ``float``, array-like
        or ``Quantity``.

    :param P:
        Orbital period(s) (if not ``Quantity``, assumed to be in days)
    :type P:
        ``float``, array-like or ``Quantity``.

    :param ecc: (``float`` or array-like, optional)
        Eccentricities.

    :param n: (optional)
        Number of instances to simulate.  If not provided, then this number
        will be the length of ``M2`` (or ``P``) provided.

    :param mean_anomaly: (optional)
        Mean anomalies of orbits.  Usually this will just be set randomly,
        but can be provided to initialize a particular state (e.g., when
        restoring an :class:`OrbitPopulation` object from a saved state).

    :param obsx, obsy, obsz: (optional)
        "Observer" positions to define coordinates.  Will be set randomly
        if not provided.

    :param obspos: (optional)
        "Observer" positions may be set with a ``SkyCoord`` object (replaces
        obsx, obsy, obsz)
    :type obspos:
        :class:`astropy.coordinates.SkyCoord`

    """

    def __init__(self,M1,M2,P,ecc=0,n=None,
                 mean_anomaly=None,obsx=None,obsy=None,obsz=None,
                 obspos=None):

        if type(M1) != Quantity:
            M1 = Quantity(M1, unit='M_sun')
        if type(M2) != Quantity:
            M2 = Quantity(M2, unit='M_sun')
        if type(P) != Quantity:
            P = Quantity(P, unit='day')

        if n is None:
            if M2.size==1:
                n = P.size
            else:
                n = M2.size

        self.M1 = M1
        self.M2 = M2

        self.N = n

        self.P = P

        if np.size(ecc) == 1:
            ecc = np.ones(n)*ecc

        self.ecc = ecc

        mred = M1*M2/(M1+M2)
        self.semimajor = semimajor(P,mred)   #AU
        self.mred = mred

        if mean_anomaly is None:
            M = rand.uniform(0,2*np.pi,size=n)
        else:
            M = mean_anomaly

        self.M = M

        #coordinates of random observers
        if obspos is None:
            if obsx is None:
                self.obspos = random_spherepos(n)
            else:
                self.obspos = SkyCoord(obsx,obsy,obsz,representation='cartesian')
        else:
            self.obspos = obspos

        #get positions, velocities relative to M1
        position,velocity = orbit_posvel(self.M,self.ecc,self.semimajor.value,
                                            self.mred.value,
                                            self.obspos)

        self.position = position
        self.velocity = velocity

    def __add__(self, other):
        if type(self) != type(other):
            raise TypeError('Can only add like types of OrbitPopulation')

        newdf = pd.concat((self.dataframe, other.dataframe))

        return OrbitPopulation_FromDF(newdf)

    def __hash__(self):
        return hashdf(self.dataframe)
    
    @property
    def Rsky(self):
        """
        Sky separation of stars, in projected AU. 
        """
        return np.sqrt(self.position.x**2 + self.position.y**2)

    @property
    def RV(self):
        """
        Relative radial velocities of two stars
        """
        return self.velocity.z

    @property
    def RV_com1(self):
        """
        RVs of star 1 relative to center-of-mass
        """
        return self.RV * (self.M2 / (self.M1 + self.M2))

    @property
    def RV_com2(self):
        """
        RVs of star 2 relative to center-of-mass
        """
        return -self.RV * (self.M1 / (self.M1 + self.M2))
    
    def dRV(self,dt,com=False):
        """Change in RV of star 1 for time separation dt (default=days)

        :param dt: 
            Time separation for which to compute RV change.  If not a ``Quantity``,
            then assumed to be in days. 
        :type dt:
            float, array-like, or ``Quantity``

        :param com: (``bool``, optional)
            If ``True``, then return dRV of star 1 in center-of-mass frame.

        :return dRV:
            Change in radial velocity over time ``dt``.

        """
        if type(dt) != Quantity:
            dt *= u.day

        mean_motions = np.sqrt(G*(self.mred)*MSUN/(self.semimajor*AU)**3)
        mean_motions = np.sqrt(const.G*(self.mred)/(self.semimajor)**3)

        newM = self.M + mean_motions * dt
        pos,vel = orbit_posvel(newM,self.ecc,self.semimajor.value,
                               self.mred.value,
                               self.obspos)
        
        if com:
            return (vel.z - self.RV) * (self.M2 / (self.M1 + self.M2))
        else:
            return vel.z-self.RV

    def RV_timeseries(self,ts,recalc=False):
        """
        Radial Velocity time series for star 1 at given times ts.

        :param ts:
            Times.  If not ``Quantity``, assumed to be in days.
        :type ts:
            array-like or ``Quantity``

        :param recalc: (optional)
            If ``False``, then if called with the exact same ``ts``
            as last call, it will return cached calculation.

        """
        if type(ts) != Quantity:
            ts *= u.day

        if not recalc and hasattr(self,'RV_measurements'):
            if (ts == self.ts).all():
                return self._RV_measurements
            else:
                pass
            
        RVs = Quantity(np.zeros((len(ts),self.N)),unit='km/s')
        for i,t in enumerate(ts):
            RVs[i,:] = self.dRV(t,com=True)
        self._RV_measurements = RVs
        self.ts = ts
        return RVs

    @property
    def dataframe(self):
        """
        Summary DataFrame of OrbitPopulation

        Used to save/restore state.
        """
        if not hasattr(self,'_dataframe'):
            obspos = self.obspos.represent_as('cartesian')
            obsx, obsy, obsz = (obspos.x,obspos.y,obspos.z)
            df = pd.DataFrame({'M1':self.M1,
                               'M2':self.M2,
                               'P':self.P,
                               'ecc':self.ecc,
                               'mean_anomaly':self.M,
                               'obsx':obsx,
                               'obsy':obsy,
                               'obsz':obsz})
            self._dataframe = df
        return self._dataframe


    def scatterplot(self,fig=None,figsize=(7,7),ms=0.5,
                    rmax=None,log=False,**kwargs):
        """
        Makes a scatter plot of projected X-Y sky separation

        :param fig: (optional)
            Passed to :func:`plotutils.setfig`

        :param figsize: (optional)
            Size of figure (in).

        :param ms: (optional)
            Marker size

        :param rmax: (optional)
            Maximum projected radius to plot.

        :param log: (optional)
            Whether to plot with log scale.

        :param **kwargs:
            Additional keyword arguments passed to ``plt.plot``.

        """
        setfig(fig,figsize=figsize)
        plt.plot(self.position.x.value,self.position.y.value,'o',ms=ms,**kwargs)
        plt.xlabel('projected separation [AU]')
        plt.ylabel('projected separation [AU]')
        if rmax is not None:
            plt.xlim((-rmax,rmax))
            plt.ylim((-rmax,rmax))
        if log:
            plt.xscale('log')
            plt.yscale('log')
    
    def save_hdf(self,filename,path=''):
        """
        Saves all relevant data to .h5 file; so state can be restored.
        """
        self.dataframe.to_hdf(filename,'{}/df'.format(path))

    @classmethod
    def from_df(cls, df):
        """Creates an OrbitPopulation from a DataFrame.

        :param df:
            :class:`pandas.DataFrame` object.  Must contain the following
            columns: ``['M1','M2','P','ecc','mean_anomaly','obsx','obsy','obsz']``,
            i.e., as what is accessed via :attr:`OrbitPopulation.dataframe`.

        :return:
            :class:`OrbitPopulation`.
        """
        return cls(df['M1'], df['M2'], df['P'],
                   ecc=df['ecc'], mean_anomaly=df['mean_anomaly'],
                   obsx=df['obsx'], obsy=df['obsy'], obsz=df['obsz'])
        
    @classmethod
    def load_hdf(cls, filename, path=''):
        """Loads OrbitPopulation from HDF file.

        :param filename:
            HDF file

        :param path:
            Path within HDF file store where :class:`OrbitPopulation` is saved.
        """
        df = pd.read_hdf(filename,'{}/df'.format(path), autoclose=True)
        return cls.from_df(df)
        

class BinaryGrid(OrbitPopulation):
    def __init__(self, M1, qmin=0.1, qmax=1, Pmin=0.5, Pmax=365, N=1e5, logP=True, eccfn=None):
        """A grid of companions to primary, in mass ratio and period space.


        :param M1:
            Primary mass [solar masses].
        :type M1:
            ``float``

        :param qmin,qmax: (optional)
            Minimum and maximum mass ratios.

        :param Pmin,Pmax: (optional)
            Min/max periods in days.

        :param N: (optional)
            Total number of simulations.  Default = 10^5.

        :param logP: (optional)
            Whether to grid in log-period.  If ``False``, then linear.

        :param eccfn: (optional)
            Function that returns eccentricity as a function of period.
            If ``None``, then eccentricity will be zero.
        :type eccfn:
            callable

        """
        M1s = np.ones(N)*M1
        M2s = (rand.random(size=N)*(qmax-qmin) + qmin)*M1s
        if logP:
            Ps = 10**(rand.random(size=N)*((np.log10(Pmax) - np.log10(Pmin))) + np.log10(Pmin))
        else:
            Ps = rand.random(size=N)*(Pmax - Pmin) + Pmin

        
            
        if eccfn is None:
            eccs = 0
        else:
            eccs = eccfn(Ps)

        self.eccfn = eccfn

        OrbitPopulation.__init__(self,M1s,M2s,Ps,ecc=eccs)

    def RV_RMSgrid(self,ts,res=20,mres=None,Pres=None,
                   conf=0.95,measured_rms=None,drv=0,
                   plot=True,fig=None,contour=True,sigma=1):
        """Writes a grid of RV RMS values, assuming observations at given times.

        Caveat Emptor: Written a long time ago, and 
        hasn't really been tested. 

        :param ts:
            Times of observations

        :param res, mres, Pres: (optional)
            Resolution of grids.  ``res`` relates to both mass and period;
            otherwise ``mres`` and ``Pres`` can be set separately.

        :param conf: (optional)
            Confidence at which to exclude regions.  Used if ``measured_rms``
            is ``None``.

        :param measured_rms: (optional)
            Measured RV RMS, if applicable [not sure exactly how this is used]

        :param drv: (optional)
            Uncertainties in RV to simulate.

        :param plot: (optional)
            Whether to plot result.

        :param fig: (optional)
            Passed to :func:`plotutils.setfig`.

        :param contour: (optional)
            Whether to plot contours.

        :param sigma: (optional)
            Level at which to exclude, based on ``measured_rms``.

        """
        RVs = self.RV_timeseries(ts)
        RVs += rand.normal(size=np.size(RVs)).reshape(RVs.shape)*drv
        rms = RVs.std(axis=0)

        if mres is None:
            mres = res
        if Pres is None:
            Pres = res

        mbins = np.linspace(self.M2.min(),self.M2.max(),mres+1)
        Pbins = np.logspace(np.log10(self.P.min()),np.log10(self.P.max()),Pres+1)
        logPbins = np.log10(Pbins)

        mbin_centers = (mbins[:-1] + mbins[1:])/2.
        logPbin_centers = (logPbins[:-1] + logPbins[1:])/2.

        minds = np.digitize(self.M2,mbins)
        Pinds = np.digitize(self.P,Pbins)

        pctiles = np.zeros((mres,Pres))
        ns = np.zeros((mres,Pres))

        for i in np.arange(mres):
            for j in np.arange(Pres):
                w = np.where((minds==i+1) & (Pinds==j+1))
                these = rms[w]
                n = size(these)
                ns[i,j] = n
                if measured_rms is not None:
                    pctiles[i,j] = (these > sigma*measured_rms).sum()/float(n)
                else:
                    inds = np.argsort(these)
                    pctiles[i,j] = these[inds][int((1-conf)*n)]

        Ms,logPs = np.meshgrid(mbin_centers,logPbin_centers)

        if plot:
            setfig(fig)

            if contour:
                mbin_centers = (mbins[:-1] + mbins[1:])/2.
                logPbins = np.log10(Pbins)
                logPbin_centers = (logPbins[:-1] + logPbins[1:])/2.
                if measured_rms is not None:
                    levels = [0.68,0.95,0.99]
                else:
                    levels = np.arange(0,20,2)
                c = plt.contour(logPbin_centers,mbin_centers,pctiles,levels=levels,colors='k')
                plt.clabel(c, fontsize=10, inline=1)
                
            else:
                extent = [np.log10(self.P.min()),np.log10(self.P.max()),self.M2.min(),self.M2.max()]
                im = plt.imshow(pctiles,cmap='Greys',extent=extent,aspect='auto')

                fig = plt.gcf()
                ax = plt.gca()


                if measured_rms is None:
                    cbarticks = np.arange(0,21,2)
                else:
                    cbarticks = np.arange(0,1.01,0.1)
                cbar = fig.colorbar(im, ticks=cbarticks)

            plt.xlabel('Log P')
            plt.ylabel('M2')

        return mbins,Pbins,pctiles,ns
            



import logging

try:
    import numpy as np
    import numpy.random as rand

    from astropy import units as u
    from astropy.coordinates import SkyCoord,Angle
    from astropy.units.quantity import Quantity
    from astropy import constants as const
    MSUN = const.M_sun.cgs.value
    AU = const.au.cgs.value
    DAY = 86400
    G = const.G.cgs.value
except ImportError:
    np,rand = (None, None)
    u, SkyCoord, Angle, Quantity, Const = (None, None, None, None, None)
    MSUN, AU, DAY, G = (None, None, None, None)

    
from .kepler import Efn #

def semimajor(P,M):
    """
    P, M can be ``Quantity`` objects; otherwise default to day, M_sun
    """
    if type(P) != Quantity:
        P = P*u.day
    if type(M) != Quantity:
        M = M*u.M_sun
    a = ((P/2/np.pi)**2*const.G*M)**(1./3)
    return a.to(u.AU)

def random_spherepos(n):
    """
    Returns SkyCoord object with n positions randomly oriented on the unit sphere.

    :param n: (``int``)
        Number of positions desired.

    :return c:
        ``astropy.coordinates.SkyCoord`` object with random positions 
    """
    signs = np.sign(rand.uniform(-1,1,size=n))
    thetas = Angle(np.arccos(rand.uniform(size=n)*signs),unit=u.rad) #random b/w 0 and 180
    phis = Angle(rand.uniform(0,2*np.pi,size=n),unit=u.rad)
    c = SkyCoord(phis,thetas,1,representation='physicsspherical')
    return c

def orbitproject(x,y,inc,phi=0,psi=0):
    """
    Transform x,y planar coordinates into observer's coordinate frame.

    ``x,y`` are coordinates in ``z=0`` plane (plane of the orbit)

    observer is at ``(inc, phi)`` on celestial sphere (angles in radians);
    ``psi`` is orientation of final ``x-y`` axes about the ``(inc,phi)`` vector.

    Returns ``x,y,z`` values in observer's coordinate frame, where
    ``x,y`` are now plane-of-sky coordinates and ``z`` is along the line of sight.

    :param x,y: (``float`` or array-like)
        Coordinates to transform.

    :param inc: (``float`` or array-like)
        Polar angle(s) of observer (where ``inc=0`` corresponds to north pole
        of original ``x-y`` plane).  This angle is the same as standard "inclination."

    :param phi: (``float`` or array-like, optional)
        Azimuthal angle of observer around ``z`` -axis

    :param psi: (``float`` or array-like, optional)
        Orientation of final observer coordinate frame (azimuthal around
        ``(inc,phi)`` vector.

    :return x,y,z: (``ndarray``)
        Coordinates in observers' frames.  ``x,y`` in "plane of sky" and ``z``
        along line of sight.
    """

    x2 = x*np.cos(phi) + y*np.sin(phi)
    y2 = -x*np.sin(phi) + y*np.cos(phi)
    z2 = y2*np.sin(inc)
    y2 = y2*np.cos(inc)

    xf = x2*np.cos(psi) - y2*np.sin(psi)
    yf = x2*np.sin(psi) + y2*np.cos(psi)

    return (xf,yf,z2)

def orbit_posvel(Ms,eccs,semimajors,mreds,obspos=None):
    """
    Returns positions in projected AU and velocities in km/s for given mean anomalies.

    Returns 3-D positions and velocities as SkyCoord objects, in
    "observer" reference frame.  Uses
    :func:`kepler.Efn` to calculate eccentric anomalies using interpolation.    

    :param Ms,eccs,semimajors,mreds: (``float`` or array-like)
        Mean anomalies, eccentricities, semimajor axes [AU], reduced masses [Msun].

    :param obspos: (``None``, ``(x,y,z)`` ``tuple`` or ``SkyCoord`` object)
        Locations of observers for which to return coordinates.
        If ``None`` then populate randomly on sphere.  If ``(x,y,z)`` or
        ``SkyCoord`` object provided, then use those.

    :returns pos,vel:
        ``SkyCoord`` Objects representing the positions and velocities,
        the coordinates
        of which are ``Quantity`` objects that have units.  Positions are in
        projected AU and velocities in km/s.
        
    """

    Es = Efn(Ms,eccs) #eccentric anomalies by interpolation

    rs = semimajors*(1-eccs*np.cos(Es))
    nus = 2 * np.arctan2(np.sqrt(1+eccs)*np.sin(Es/2),np.sqrt(1-eccs)*np.cos(Es/2))

    xs = semimajors*(np.cos(Es) - eccs)         #AU
    ys = semimajors*np.sqrt(1-eccs**2)*np.sin(Es)  #AU
    
    Edots = np.sqrt(G*mreds*MSUN/(semimajors*AU)**3)/(1-eccs*np.cos(Es))
        
    xdots = -semimajors*AU*np.sin(Es)*Edots/1e5  #km/s
    ydots = semimajors*AU*np.sqrt(1-eccs**2)*np.cos(Es)*Edots/1e5 # km/s
        
    n = np.size(xs)

    orbpos = SkyCoord(xs,ys,0*u.AU,representation='cartesian',unit='AU')
    orbvel = SkyCoord(xdots,ydots,0*u.km/u.s,representation='cartesian',unit='km/s')
    if obspos is None:
        obspos = random_spherepos(n) #observer position
    if type(obspos) == type((1,2,3)):
        obspos = SkyCoord(obspos[0],obspos[1],obspos[2],
                          representation='cartesian').represent_as('physicsspherical')

    if not hasattr(obspos,'theta'): #if obspos not physics spherical, make it 
        obspos = obspos.represent_as('physicsspherical')
        
    #random orientation of the sky 'x-y' coordinates
    psi = rand.random(n)*2*np.pi  

    #transform positions and velocities into observer coordinates
    x,y,z = orbitproject(orbpos.x,orbpos.y,obspos.theta,obspos.phi,psi)
    vx,vy,vz = orbitproject(orbvel.x,orbvel.y,obspos.theta,obspos.phi,psi)

    return (SkyCoord(x,y,z,representation='cartesian'),
            SkyCoord(vx,vy,vz,representation='cartesian')) 



from __future__ import print_function, division

import os
import logging
import pkg_resources

#test if building documentation on readthedocs.org
on_rtd = False

try:
    import numpy as np
    import numpy.random as rand
    from scipy.optimize import leastsq
    from scipy.ndimage import convolve1d
    from scipy.interpolate import LinearNDInterpolator as interpnd
except ImportError:
    on_rtd = True
    np, rand, leastsq, convolve1d, interpnd = (None, None, None, None, None)
    

from .orbits.kepler import Efn
from .stars.utils import rochelobe, withinroche, semimajor

if not on_rtd:
    from vespa_transitutils import find_eclipse
    from vespa_transitutils import traptransit, traptransit_resid
    
    import emcee
else:
    find_eclipse, traptransit, traptransit_resid = (None, None, None)
    emcee = None

if not on_rtd:
    import astropy.constants as const
    AU = const.au.cgs.value
    RSUN = const.R_sun.cgs.value
    MSUN = const.M_sun.cgs.value
    REARTH = const.R_earth.cgs.value
    MEARTH = const.M_earth.cgs.value
    DAY = 86400

    DATAFOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))

    LDDATA = np.recfromtxt('{}/keplerld.dat'.format(DATAFOLDER),names=True)
    LDOK = ((LDDATA.teff < 10000) & (LDDATA.logg > 2.0) & (LDDATA.feh > -2))
    LDPOINTS = np.array([LDDATA.teff[LDOK],LDDATA.logg[LDOK]]).T
    U1FN = interpnd(LDPOINTS,LDDATA.u1[LDOK])
    U2FN = interpnd(LDPOINTS,LDDATA.u2[LDOK])
else:
    const, AU, RSUN, MSUN = (None, None, None, None)
    REARTH, MEARTH, DAY = (None, None, None)
    DATAFOLDER = None
    LDDATA, LDOK, LDPOINTS, U1FN, U2FN = (None, None, None, None, None)
    

def ldcoeffs(teff,logg=4.5,feh=0):
    """
    Returns limb-darkening coefficients in Kepler band.
    """
    teffs = np.atleast_1d(teff)
    loggs = np.atleast_1d(logg)

    Tmin,Tmax = (LDPOINTS[:,0].min(),LDPOINTS[:,0].max())
    gmin,gmax = (LDPOINTS[:,1].min(),LDPOINTS[:,1].max())

    teffs[(teffs < Tmin)] = Tmin + 1
    teffs[(teffs > Tmax)] = Tmax - 1
    loggs[(loggs < gmin)] = gmin + 0.01
    loggs[(loggs > gmax)] = gmax - 0.01

    u1,u2 = (U1FN(teffs,loggs),U2FN(teffs,loggs))
    return u1,u2

"""
def correct_fs(fs):
    fflat = fs.ravel().copy()
    wbad = np.where(fflat > 1)[0]     

    #identify lowest and highest index of valid flux
    ilowest=0
    while fflat[ilowest] > 1:
        ilowest += 1
    ihighest = len(fflat)-1
    while fflat[ihighest] > 1:
        ihighest -= 1

    wlo = wbad - 1
    whi = wbad + 1


    #find places where wlo index is still in wbad
    ilo = np.searchsorted(wbad,wlo)
    mask = wbad[ilo]==wlo
    while np.any(mask):
        wlo[mask] -= 1
        ilo = np.searchsorted(wbad,wlo)
        mask = wbad[ilo]==wlo
    ihi = np.searchsorted(wbad,whi)
    ihi = np.clip(ihi,0,len(wbad)-1) #make sure no IndexError
    mask = wbad[ihi]==whi

    while np.any(mask):
        whi[mask] += 1
        ihi = np.searchsorted(wbad,whi)
        ihi = np.clip(ihi,0,len(wbad)-1) #make sure no IndexError
        mask = wbad[ihi]==whi
    
    wlo = np.clip(wlo,ilowest,ihighest)
    whi = np.clip(whi,ilowest,ihighest)

    fflat[wbad] = (fflat[whi] + fflat[wlo])/2. #slightly kludge-y, esp. if there are consecutive bad vals
    return fflat.reshape(fs.shape)
"""

class MAInterpolationFunction(object):
    """
    Object enabling fast, vectorized evaluations of Mandel-Agol transit model.

    Interpolates on pre-defined grid calculating Mandel & Agol (2002)
    and Agol & Eastman (2008) calculations.

    This object is generally used as follows::

        >>> import numpy as np
        >>> from vespa import MAInterpolationFunction
        >>> mafn = MAInterpolationFunction() #takes a few seconds
        >>> ps = 0.1 # radius ratio; can be float or array-like
        >>> zs = np.abs(np.linspace(-1,1,1000)) #impact parameters
        >>> fs = mafn(ps, zs) # relative flux

    Even cooler, it can be called with different-sized arrays for
    radius ratio and impact parameter, in which case it returns a
    flux array of shape ``(nps, nzs)``.  This is clearly awesome
    for generating populations of eclipses::

        >>> ps = np.linspace(0.01,0.1,100) # radius ratios
        >>> zs = np.abs(np.linspace(-1,1,1000)) #impact parameters
        >>> fs = mafn(ps, zs)
        >>> fs.shape
            (100, 1000)

    It can also be called with different limb darkening parameters,
    in which case arrays of ``u1`` and ``u2`` should be the third
    and fourth argument, after ``ps`` and ``zs``, with the same shape
    as ``ps`` (radius ratios).  
            
    :param u1,u2: (optional)
        Default quadratic limb darkening parameters. Setting
        these only enables faster evaluation; you can always call with
        different values.

    :param pmin,pmax,nps,nzs,zmax: (optional)
        Parameters describing grid in p and z.
        
    """
    def __init__(self,u1=0.394,u2=0.261,pmin=0.007,pmax=2,nps=200,nzs=200,zmax=None):
        self.u1 = u1
        self.u2 = u2
        self.pmin = pmin
        self.pmax = pmax
        if zmax is None:
            zmax = 1+pmax
        self.zmax = zmax
        self.nps = nps

        ps = np.logspace(np.log10(pmin),np.log10(pmax),nps)
        if pmax < 0.5:
            zs = np.concatenate([np.array([0]),ps-1e-10,ps,np.arange(pmax,1-pmax,0.01),
                              np.arange(1-pmax,zmax,0.005)])
        elif pmax < 1:
            zs = np.concatenate([np.array([0]),ps-1e-10,ps,np.arange(1-pmax,zmax,0.005)])
        else:
            zs = np.concatenate([np.array([0]),ps-1e-10,ps,np.arange(pmax,zmax,0.005)])

        self.nzs = np.size(zs)
        #zs = linspace(0,zmax,nzs)
        #zs = concatenate([zs,ps,ps+1e-10])

        mu0s = np.zeros((np.size(ps),np.size(zs)))
        lambdads = np.zeros((np.size(ps),np.size(zs)))
        etads = np.zeros((np.size(ps),np.size(zs)))
        fs = np.zeros((np.size(ps),np.size(zs)))
        for i,p0 in enumerate(ps):
            f,res = occultquad(zs,u1,u2,p0,return_components=True)
            mu0s[i,:] = res[0]
            lambdads[i,:] = res[1]
            etads[i,:] = res[2]
            fs[i,:] = f
        P,Z = np.meshgrid(ps,zs)
        points = np.array([P.ravel(),Z.ravel()]).T
        self.mu0 = interpnd(points,mu0s.T.ravel())
        
        ##need to make two interpolation functions for lambdad 
        ## b/c it's strongly discontinuous at z=p
        mask = (Z<P)
        pointmask = points[:,1] < points[:,0]

        w1 = np.where(mask)
        w2 = np.where(~mask)
        wp1 = np.where(pointmask)
        wp2 = np.where(~pointmask)

        self.lambdad1 = interpnd(points[wp1],lambdads.T[w1].ravel())
        self.lambdad2 = interpnd(points[wp2],lambdads.T[w2].ravel())
        def lambdad(p,z):
            #where p and z are exactly equal, this will return nan....
            p = np.atleast_1d(p)
            z = np.atleast_1d(z)
            l1 = self.lambdad1(p,z)
            l2 = self.lambdad2(p,z)
            bad1 = np.isnan(l1)
            l1[np.where(bad1)]=0
            l2[np.where(~bad1)]=0
            return l1*~bad1 + l2*bad1
        self.lambdad = lambdad
        
        #self.lambdad = interpnd(points,lambdads.T.ravel())
        self.etad = interpnd(points,etads.T.ravel())        
        self.fn = interpnd(points,fs.T.ravel())

    def __call__(self,ps,zs,u1=.394,u2=0.261,force_broadcast=False):
        """  returns array of fluxes; if ps and zs aren't the same shape, then returns array of 
        shape (nps, nzs)
        """
        #return self.fn(ps,zs)

        if np.size(ps)>1 and (np.size(ps)!=np.size(zs) or force_broadcast):
            P = ps[:,None]
            if np.size(u1)>1 or np.size(u2)>1:
                if u1.shape != ps.shape or u2.shape != ps.shape:
                    raise ValueError('limb darkening coefficients must be same size as ps')
                U1 = u1[:,None]
                U2 = u2[:,None]
            else:
                U1 = u1
                U2 = u2
        else:
            P = ps
            U1 = u1
            U2 = u2
            
        if np.size(u1)>1 or np.any(u1 != self.u1) or np.any(u2 != self.u2):
            mu0 = self.mu0(P,zs)
            lambdad = self.lambdad(P,zs)
            etad = self.etad(P,zs)
            fs = 1. - ((1-U1-2*U2)*(1-mu0) + (U1+2*U2)*(lambdad+2./3*(P > zs)) + U2*etad)/(1.-U1/3.-U2/6.)
            
            #if fix:
            #    fs = correct_fs(fs)
        else:
            fs = self.fn(P,zs)
            
        return fs

def impact_parameter(a, R, inc, ecc=0, w=0, return_occ=False):
    """a in AU, R in Rsun, inc & w in radians
    """
    b_tra = a*AU*np.cos(inc)/(R*RSUN) * (1-ecc**2)/(1 + ecc*np.sin(w))

    if return_occ:
        b_tra = a*AU*np.cos(inc)/(R*RSUN) * (1-ecc**2)/(1 - ecc*np.sin(w))
        return b_tra, b_occ
    else:
        return b_tra

def transit_T14(P,Rp,Rs=1,b=0,Ms=1,ecc=0,w=0):
    """P in days, Rp in Earth radii, Rs in Solar radii, b=impact parameter, Ms Solar masses. Returns T14 in hours. w in deg.
    """
    a = semimajor(P,Ms)*AU
    k = Rp*REARTH/(Rs*RSUN)
    inc = np.pi/2 - b*RSUN/a
    return  P*DAY/np.pi*np.arcsin(Rs*RSUN/a * np.sqrt((1+k)**2 - b**2)/np.sin(inc)) *\
        np.sqrt(1-ecc**2)/(1+ecc*np.sin(w*np.pi/180)) / 3600.

def transit_T23(P,Rp,Rs=1,b=0,Ms=1,ecc=0,w=0):
    a = semimajor(P,Ms)*AU
    k = Rp*REARTH/(Rs*RSUN)
    inc = np.pi/2 - b*RSUN/a

    return P*DAY/np.pi*np.arcsin(Rs*RSUN/a * np.sqrt((1-k)**2 - b**2)/np.sin(inc)) *\
        np.sqrt(1-ecc**2)/(1+ecc*np.sin(w*pi/180)) / 3600.#*24*60    

def eclipse_depth(mafn,Rp,Rs,b,u1=0.394,u2=0.261,max_only=False,npts=100,force_1d=False):
    """ Calculates average (or max) eclipse depth

    ***why does b>1 take so freaking long?...
    """
    k = Rp*REARTH/(Rs*RSUN)
    
    if max_only:
        return 1 - mafn(k,b,u1,u2)

    if np.size(b) == 1:
        x = np.linspace(0,np.sqrt(1-b**2),npts)
        y = b
        zs = np.sqrt(x**2 + y**2)
        fs = mafn(k,zs,u1,u2) # returns array of shape (nks,nzs)
        depth = 1-fs
    else:
        xmax = np.sqrt(1-b**2)
        x = np.linspace(0,1,npts)*xmax[:,Nones]
        y = b[:,None]
        zs = np.sqrt(x**2 + y**2)
        fs = mafn(k,zs.ravel(),u1,u2)
        if not force_1d:
            fs = fs.reshape(size(k),*zs.shape)
        depth = 1-fs
        
    meandepth = np.squeeze(depth.mean(axis=depth.ndim-1))
    
    return meandepth  #array of average depths, shape (nks,nbs)

def minimum_inclination(P,M1,M2,R1,R2):
    """
    Returns the minimum inclination at which two bodies from two given sets eclipse

    Only counts systems not within each other's Roche radius
    
    :param P:
        Orbital periods.

    :param M1,M2,R1,R2:
        Masses and radii of primary and secondary stars.  
    """
    P,M1,M2,R1,R2 = (np.atleast_1d(P),
                     np.atleast_1d(M1),
                     np.atleast_1d(M2),
                     np.atleast_1d(R1),
                     np.atleast_1d(R2))
    semimajors = semimajor(P,M1+M2)
    rads = ((R1+R2)*RSUN/(semimajors*AU))
    ok = (~np.isnan(rads) & ~withinroche(semimajors,M1,R1,M2,R2))
    if ok.sum() == 0:
        logging.error('P: {}'.format(P))
        logging.error('M1: {}'.format(M1))
        logging.error('M2: {}'.format(M2))
        logging.error('R1: {}'.format(R1))
        logging.error('R2: {}'.format(R2))
        if np.all(withinroche(semimajors,M1,R1,M2,R2)):
            raise AllWithinRocheError('All simulated systems within Roche lobe')
        else:
            raise EmptyPopulationError('no valid systems! (see above)')
    mininc = np.arccos(rads[ok].max())*180/np.pi
    return mininc

def a_over_Rs(P,R2,M2,M1=1,R1=1,planet=True):
    """
    Returns a/Rs for given parameters.
    """
    if planet:
        M2 *= REARTH/RSUN
        R2 *= MEARTH/MSUN
    return semimajor(P,M1+M2)*AU/(R1*RSUN)

def eclipse_tz(P,b,aR,ecc=0,w=0,npts=200,width=1.5,sec=False,dt=1,approx=False,new=True):
    """Returns ts and zs for an eclipse (npts points right around the eclipse)

    
    :param P,b,aR:
        Period, impact parameter, a/Rstar.

    :param ecc,w:
        Eccentricity, argument of periapse.

    :param npts:
        Number of points in transit to return.

    :param width:
        How much to go out of transit. 1.5 is a good choice.

    :param sec:
        Whether to return the values relevant to the secondary occultation
        rather than primary eclipse.

    :param dt:
        Spacing of simulated data points, in minutes.

    :param approx:
        Whether to use the approximate expressions to find the mean
        anomalies.  Default is ``False`` (exact calculation).

    :param new:
        Meaningless.

    :return ts,zs:
        Times from mid-transit (in days) and impact parameters.
        
    """
    if sec:
        eccfactor = np.sqrt(1-ecc**2)/(1-ecc*np.sin(w*np.pi/180))
    else:
        eccfactor = np.sqrt(1-ecc**2)/(1+ecc*np.sin(w*np.pi/180))
    if eccfactor < 1:
        width /= eccfactor
        #if width > 5:
        #    width = 5
        
    if new:
        #finding the mean anomaly boundaries of the eclipse

        #shoot for at least 100 pts in transit
        n = 100 * (np.pi * aR)

        Ms = np.linspace(-np.pi,np.pi,n)
        if ecc != 0:
            Es = Efn(Ms,ecc) #eccentric anomalies
        else:
            Es = Ms
        zs,in_eclipse = find_eclipse(Es,b,aR,ecc,w,width,sec)

        if in_eclipse.sum() < 2:
            logging.debug('no eclipse because fewer than 2 points '+
                          'in eclipse. [Es,b,aR,ecc,w,width,sec] = ' +
                          '{}'.format([Es,b,aR,ecc,w,width,sec]))
            raise NoEclipseError

        wecl = np.where(in_eclipse)
        subMs = Ms[wecl]

        dMs = subMs[1:] - subMs[:-1]

        if np.any(subMs < 0) and dMs.max()>1: #if there's a discontinuous wrap-around...
            subMs[np.where(subMs < 0)] += 2*np.pi


        #logging.debug('subMs: {}'.format(subMs))


        minM,maxM = (subMs.min(),subMs.max())
        #logging.debug('minM: {}, maxM: {}'.format(minM,maxM))


        #dM = 2*np.pi*dt/(P*24*60)   #the spacing in mean anomaly that corresponds to dt (minutes)
        #Ms = np.arange(minM,maxM+dM,dM)

        Ms = np.linspace(minM, maxM, npts) #npts desired in transit, rather than using dt

        if ecc != 0:
            Es = Efn(Ms,ecc) #eccentric anomalies
        else:
            Es = Ms

        zs,in_eclipse = find_eclipse(Es,b,aR,ecc,w,width,sec)

        Mcenter = Ms[zs.argmin()]
        phs = (Ms - Mcenter) / (2*np.pi)
        ts = phs*P
        
        #logging.debug('{} in-transit points simulated'.format(len(ts)))

        return ts,zs
    
    if not approx:
        if sec:
            inc = np.arccos(b/aR*(1-ecc*np.sin(w*np.pi/180))/(1-ecc**2))
        else:
            inc = np.arccos(b/aR*(1+ecc*np.sin(w*np.pi/180))/(1-ecc**2))

        Ms = np.linspace(-np.pi,np.pi,2e3) #mean anomalies around whole orbit
        if ecc != 0:
            Es = Efn(Ms,ecc) #eccentric anomalies
            nus = 2 * np.arctan2(np.sqrt(1+ecc)*np.sin(Es/2),
                                 np.sqrt(1-ecc)*np.cos(Es/2)) #true anomalies
        else:
            nus = Ms

        r = aR*(1-ecc**2)/(1+ecc*np.cos(nus))  #secondary distance from primary in units of R1

        X = -r*np.cos(w*np.pi/180 + nus)
        Y = -r*np.sin(w*np.pi/180 + nus)*np.cos(inc)
        rsky = np.sqrt(X**2 + Y**2)

        if not sec:
            inds = np.where((np.sin(nus + w*np.pi/180) > 0) & (rsky < width))  #where "front half" of orbit and w/in width
        if sec:
            inds = np.where((np.sin(nus + w*np.pi/180) < 0) & (rsky < width))  #where "front half" of orbit and w/in width
        subMs = Ms[inds].copy()

        if np.any((subMs[1:]-subMs[:-1]) > np.pi):
            subMs[np.where(subMs < 0)] += 2*np.pi

        if np.size(subMs)<2:
            logging.debug('no eclipse because fewer than 2 points '+
                          'in subMs.  Look into this...')                              
            logging.error(subMs)
            raise NoEclipseError

        minM,maxM = (subMs.min(),subMs.max())
        dM = 2*np.pi*dt/(P*24*60)   #the spacing in mean anomaly that corresponds to dt (minutes)
        Ms = np.arange(minM,maxM+dM,dM)
        if ecc != 0:
            Es = Efn(Ms,ecc) #eccentric anomalies
            nus = 2 * np.arctan2(np.sqrt(1+ecc)*np.sin(Es/2),
                                 np.sqrt(1-ecc)*np.cos(Es/2)) #true anomalies
        else:
            nus = Ms
        r = aR*(1-ecc**2)/(1+ecc*np.cos(nus))
        X = -r*np.cos(w*np.pi/180 + nus)
        Y = -r*np.sin(w*np.pi/180 + nus)*np.cos(inc)
        zs = np.sqrt(X**2 + Y**2)  #rsky
    
        #center = absolute(X).argmin()
        #c = polyfit(Ms[center-1:center+2],X[center-1:center+2],1)
        #Mcenter = -c[1]/c[0]
        if not sec:
            Mcenter = Ms[np.absolute(X[np.where(np.sin(nus + w*np.pi/180) > 0)]).argmin()]
        else:
            Mcenter = Ms[np.absolute(X[np.where(np.sin(nus + w*np.pi/180) < 0)]).argmin()]
        phs = (Ms - Mcenter) / (2*np.pi)
        wmin = np.absolute(phs).argmin()
        ts = phs*P

        return ts,zs
    else:
        if sec:
            f0 = -pi/2 - (w*pi/180)
            inc = arccos(b/aR*(1-ecc*sin(w*pi/180))/(1-ecc**2))
        else:
            f0 = pi/2 - (w*pi/180)
            inc = arccos(b/aR*(1+ecc*sin(w*pi/180))/(1-ecc**2))
        fmin = -arcsin(1./aR*sqrt(width**2 - b**2)/sin(inc))
        fmax = arcsin(1./aR*sqrt(width**2 - b**2)/sin(inc))
        if isnan(fmin) or isnan(fmax):
            logging.debug('No eclipse in approximate calculation. ' +
                          'P=%.2f, b=%.3f, aR=%.2f, ecc=%0.2f, w=%.1f' % (P,b,aR,ecc,w))
            raise NoEclipseError
        fs = linspace(fmin,fmax,npts)
        if sec:
            ts = fs*P/2./pi * sqrt(1-ecc**2)/(1 - ecc*sin(w)) #approximation of constant angular velocity
        else:
            ts = fs*P/2./pi * sqrt(1-ecc**2)/(1 + ecc*sin(w)) #approximation of constant ang. vel.
        fs += f0
        rs = aR*(1-ecc**2)/(1+ecc*cos(fs))
        xs = -rs*cos(w*pi/180 + fs)
        ys = -rs*sin(w*pi/180 + fs)*cos(inc)
        zs = aR*(1-ecc**2)/(1+ecc*cos(fs))*sqrt(1-(sin(w*pi/180 + fs))**2 * (sin(inc))**2)
        return ts,zs


def eclipse_pars(P,M1,M2,R1,R2,ecc=0,inc=90,w=0,sec=False):
    """retuns p,b,aR from P,M1,M2,R1,R2,ecc,inc,w"""
    a = semimajor(P,M1+M2)
    if sec:
        b = a*AU*np.cos(inc*np.pi/180)/(R1*RSUN) * (1-ecc**2)/(1 - ecc*np.sin(w*np.pi/180))
        aR = a*AU/(R2*RSUN)
        p0 = R1/R2
    else:
        b = a*AU*np.cos(inc*np.pi/180)/(R1*RSUN) * (1-ecc**2)/(1 + ecc*np.sin(w*np.pi/180))
        aR = a*AU/(R1*RSUN)
        p0 = R2/R1
    return p0,b,aR

def eclipse(p0,b,aR,P=1,ecc=0,w=0,npts=200,MAfn=None,u1=0.394,u2=0.261,width=3,conv=False,cadence=0.020434028,frac=1,sec=False,dt=2,approx=False,new=True):
    """Returns ts, fs of simulated eclipse.


    :param p0,b,aR:
        Radius ratio, impact parameter, a/Rstar.

    :param P:
        Orbital period

    :param ecc,w: (optional)
        Eccentricity, argument of periapse.

    :param npts: (optional)
        Number of points to simulate.

    :param MAfn: (optional)
        :class:`MAInterpolationFunction` object.

    :param u1,u2: (optional)
        Quadratic limb darkening parameters.

    :param width: (optional)
        Argument defining how much out-of-transit to simulate.  3 is good.

    :param conv: (optional)
        Whether to convolve with box-car to simulate integration time.

    :param cadence: (optional)
        Cadence to simulate; default is Kepler cadence.

    :param frac: (optional)
        Fraction of total light in eclipsed object (for dilution purposes).

    :param sec: (optional)
        If ``True``, then simulate secondary occultation rather than eclipse.

    :param dt: (optional)
        Simulated spacing of theoretical data points, in minutes.

    :param approx: (optional)
        Whether to approximate solution to Kepler's equation or not.

    :param new: (optional)
        Meaningless relic [or not??].  Apparently this is in practice
        ``True`` by default.


    :return ts,fs:
        Times from mid-transit [days] and relative fluxes of simulated eclipse.

    """

    if sec:
        ts,zs = eclipse_tz(P,b/p0,aR/p0,ecc,w,npts=npts,width=(1+1/p0)*width,
                           sec=sec,dt=dt,approx=approx,new=new)
        if zs.min() > (1 + 1/p0):
            logging.debug('no eclipse because min z is greater than 1 + 1/p0: ' +
                          '[P,b/p0,aR/p0,ecc,w] = {}'.format([P,b/p0,aR/p0,ecc,w]))                         
            raise NoEclipseError
    else:
        ts,zs = eclipse_tz(P,b,aR,ecc,w,npts=npts,width=(1+p0)*width,
                           sec=sec,dt=dt,approx=approx,new=new)
        if zs.min() > (1+p0):
            logging.debug('no eclipse (secondary) because min z is greater ' +
                          'than 1 + 1/p0: ' +
                          '[P,b,aR,ecc,w] = {}'.format([P,b,aR,ecc,w]))
            raise NoEclipseError
        
    if MAfn is None:
        if sec:
            fs = occultquad(zs,u1,u2,1/p0)
        else:
            fs = occultquad(zs,u1,u2,p0)            
    else:
        if sec:
            fs = MAfn(1/p0,zs,u1,u2)
        else:
            fs = MAfn(p0,zs,u1,u2)
        fs[np.isnan(fs)] = 1.

    if conv:
        dt = ts[1]-ts[0]
        npts = np.round(cadence/dt)
        if npts % 2 == 0:
            npts += 1
        boxcar = np.ones(npts)/npts
        fs = convolve1d(fs,boxcar)
    fs = 1 - frac*(1-fs)
    return ts,fs #ts are in the same units P is given in.

def eclipse_tt(p0,b,aR,P=1,ecc=0,w=0,npts=200,MAfn=None,u1=0.394,u2=0.261,conv=False,cadence=0.020434028,frac=1,sec=False,new=True,pars0=None):
    """
    Trapezoidal parameters for simulated orbit.
    
    All arguments passed to :func:`eclipse` except the following:

    :param pars0: (optional)
        Initial guess for least-sq optimization for trapezoid parameters.

    :return dur,dep,slope:
        Best-fit duration, depth, and T/tau for eclipse shape.
    
    """
    ts,fs = eclipse(p0,b,aR,P,ecc,w,npts,MAfn,u1,u2,
                    conv=conv,cadence=cadence,frac=frac,sec=sec,new=new)
    
    #logging.debug('{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}'.format(p0,b,aR,P,ecc,w,xmax,npts,u1,u2,leastsq,conv,cadence,frac,sec,new))
    #logging.debug('ts: {} fs: {}'.format(ts,fs))

    if pars0 is None:
        depth = 1 - fs.min()
        duration = (fs < (1-0.01*depth)).sum()/float(len(fs)) * (ts[-1] - ts[0])
        tc0 = ts[fs.argmin()]
        pars0 = np.array([duration,depth,5.,tc0])
    
    dur,dep,slope,epoch = fit_traptransit(ts,fs,pars0)
    return dur,dep,slope




def occultquad(z,u1,u2,p0,return_components=False):
    """
    #### Mandel-Agol code:
    #   Python translation of IDL code.
    #   This routine computes the lightcurve for occultation of a
    #   quadratically limb-darkened source without microlensing.  Please
    #   cite Mandel & Agol (2002) and Eastman & Agol (2008) if you make use
    #   of this routine in your research.  Please report errors or bugs to
    #   jdeast@astronomy.ohio-state.edu

    .. note::

        Should probably wrap the Fortran code at some point.
        (This particular part of the code was put together awhile ago.)

    """
    z = np.atleast_1d(z)
    nz = np.size(z)
    lambdad = np.zeros(nz)
    etad = np.zeros(nz)
    lambdae = np.zeros(nz)
    omega=1.-u1/3.-u2/6.

    ## tolerance for double precision equalities
    ## special case integrations
    tol = 1e-14

    p = np.absolute(p0)
    
    z = np.where(np.absolute(p-z) < tol,p,z)
    z = np.where(np.absolute((p-1)-z) < tol,p-1.,z)
    z = np.where(np.absolute((1-p)-z) < tol,1.-p,z)
    z = np.where(z < tol,0.,z)
               
    x1=(p-z)**2.
    x2=(p+z)**2.
    x3=p**2.-z**2.
    

    def finish(p,z,u1,u2,lambdae,lambdad,etad):
        omega = 1. - u1/3. - u2/6.
        #avoid Lutz-Kelker bias
        if p0 > 0:
            #limb darkened flux
            muo1 = 1 - ((1-u1-2*u2)*lambdae+(u1+2*u2)*(lambdad+2./3*(p > z)) + u2*etad)/omega
            #uniform disk
            mu0 = 1 - lambdae
        else:
            #limb darkened flux
            muo1 = 1 + ((1-u1-2*u2)*lambdae+(u1+2*u2)*(lambdad+2./3*(p > z)) + u2*etad)/omega
            #uniform disk
            mu0 = 1 + lambdae
        if return_components:
            return muo1,(mu0,lambdad,etad)
        else:
            return muo1



    ## trivial case of no planet
    if p <= 0.:
        return finish(p,z,u1,u2,lambdae,lambdad,etad)

    ## Case 1 - the star is unocculted:
    ## only consider points with z lt 1+p
    notusedyet = np.where( z < (1. + p) )[0]
    if np.size(notusedyet) == 0:
        return finish(p,z,u1,u2,lambdae,lambdad,etad)

    # Case 11 - the  source is completely occulted:
    if p >= 1.:
        cond = z[notusedyet] <= p-1.
        occulted = np.where(cond)#,complement=notused2)
        notused2 = np.where(~cond)
        #occulted = where(z[notusedyet] <= p-1.)#,complement=notused2)
        if np.size(occulted) != 0:
            ndxuse = notusedyet[occulted]
            etad[ndxuse] = 0.5 # corrected typo in paper
            lambdae[ndxuse] = 1.
            # lambdad = 0 already
            #notused2 = where(z[notusedyet] > p-1)
            if np.size(notused2) == 0:
                return finish(p,z,u1,u2,lambdae,lambdad,etad)
            notusedyet = notusedyet[notused2]
                
    # Case 2, 7, 8 - ingress/egress (uniform disk only)
    inegressuni = np.where((z[notusedyet] >= np.absolute(1.-p)) & (z[notusedyet] < 1.+p))
    if np.size(inegressuni) != 0:
        ndxuse = notusedyet[inegressuni]
        tmp = (1.-p**2.+z[ndxuse]**2.)/2./z[ndxuse]
        tmp = np.where(tmp > 1.,1.,tmp)
        tmp = np.where(tmp < -1.,-1.,tmp)
        kap1 = np.arccos(tmp)
        tmp = (p**2.+z[ndxuse]**2-1.)/2./p/z[ndxuse]
        tmp = np.where(tmp > 1.,1.,tmp)
        tmp = np.where(tmp < -1.,-1.,tmp)
        kap0 = np.arccos(tmp)
        tmp = 4.*z[ndxuse]**2-(1.+z[ndxuse]**2-p**2)**2
        tmp = np.where(tmp < 0,0,tmp)
        lambdae[ndxuse] = (p**2*kap0+kap1 - 0.5*np.sqrt(tmp))/np.pi
        # eta_1
        etad[ndxuse] = 1./2./np.pi*(kap1+p**2*(p**2+2.*z[ndxuse]**2)*kap0- \
           (1.+5.*p**2+z[ndxuse]**2)/4.*np.sqrt((1.-x1[ndxuse])*(x2[ndxuse]-1.)))
    
    # Case 5, 6, 7 - the edge of planet lies at origin of star
    cond = z[notusedyet] == p
    ocltor = np.where(cond)#, complement=notused3)
    notused3 = np.where(~cond)
    #ocltor = where(z[notusedyet] == p)#, complement=notused3)
    t = np.where(z[notusedyet] == p)
    if np.size(ocltor) != 0:
        ndxuse = notusedyet[ocltor] 
        if p < 0.5:
            # Case 5
            q=2.*p  # corrected typo in paper (2k -> 2p)
            Ek,Kk = ellke(q)
            # lambda_4
            lambdad[ndxuse] = 1./3.+2./9./np.pi*(4.*(2.*p**2-1.)*Ek+\
                                              (1.-4.*p**2)*Kk)
            # eta_2
            etad[ndxuse] = p**2/2.*(p**2+2.*z[ndxuse]**2)        
            lambdae[ndxuse] = p**2 # uniform disk
        elif p > 0.5:
            # Case 7
            q=0.5/p # corrected typo in paper (1/2k -> 1/2p)
            Ek,Kk = ellke(q)
            # lambda_3
            lambdad[ndxuse] = 1./3.+16.*p/9./np.pi*(2.*p**2-1.)*Ek-\
                              (32.*p**4-20.*p**2+3.)/9./np.pi/p*Kk
            # etad = eta_1 already
        else:
            # Case 6
            lambdad[ndxuse] = 1./3.-4./np.pi/9.
            etad[ndxuse] = 3./32.
        #notused3 = where(z[notusedyet] != p)
        if np.size(notused3) == 0:
            return finish(p,z,u1,u2,lambdae,lambdad,etad)
        notusedyet = notusedyet[notused3]

    # Case 2, Case 8 - ingress/egress (with limb darkening)
    cond = ((z[notusedyet] > 0.5+np.absolute(p-0.5)) & \
                       (z[notusedyet] < 1.+p))  | \
                      ( (p > 0.5) & (z[notusedyet] > np.absolute(1.-p)) & \
                        (z[notusedyet] < p))
    inegress = np.where(cond)
    notused4 = np.where(~cond)
    #inegress = where( ((z[notusedyet] > 0.5+abs(p-0.5)) & \
        #(z[notusedyet] < 1.+p))  | \
        #( (p > 0.5) & (z[notusedyet] > abs(1.-p)) & \
        #(z[notusedyet] < p)) )#, complement=notused4)
    if np.size(inegress) != 0:

        ndxuse = notusedyet[inegress]
        q=np.sqrt((1.-x1[ndxuse])/(x2[ndxuse]-x1[ndxuse]))
        Ek,Kk = ellke(q)
        n=1./x1[ndxuse]-1.

        # lambda_1:
        lambdad[ndxuse]=2./9./np.pi/np.sqrt(x2[ndxuse]-x1[ndxuse])*\
                         (((1.-x2[ndxuse])*(2.*x2[ndxuse]+x1[ndxuse]-3.)-\
                           3.*x3[ndxuse]*(x2[ndxuse]-2.))*Kk+(x2[ndxuse]-\
                           x1[ndxuse])*(z[ndxuse]**2+7.*p**2-4.)*Ek-\
                          3.*x3[ndxuse]/x1[ndxuse]*ellpic_bulirsch(n,q))

        #notused4 = where( ( (z[notusedyet] <= 0.5+abs(p-0.5)) | \
        #                    (z[notusedyet] >= 1.+p) ) & ( (p <= 0.5) | \
        #                    (z[notusedyet] <= abs(1.-p)) | \
        #                    (z[notusedyet] >= p) ))
        if np.size(notused4) == 0:
            return finish(p,z,u1,u2,lambdae,lambdad,etad)
        notusedyet = notusedyet[notused4]

    # Case 3, 4, 9, 10 - planet completely inside star
    if p < 1.:
        cond = z[notusedyet] <= (1.-p)
        inside = np.where(cond)
        notused5 = np.where(~cond)
        #inside = where(z[notusedyet] <= (1.-p))#, complement=notused5)
        if np.size(inside) != 0:
            ndxuse = notusedyet[inside]

            ## eta_2
            etad[ndxuse] = p**2/2.*(p**2+2.*z[ndxuse]**2)

            ## uniform disk
            lambdae[ndxuse] = p**2

            ## Case 4 - edge of planet hits edge of star
            edge = np.where(z[ndxuse] == 1.-p)#, complement=notused6)
            if np.size(edge[0]) != 0:
                ## lambda_5
                lambdad[ndxuse[edge]] = 2./3./np.pi*np.arccos(1.-2.*p)-\
                                      4./9./np.pi*np.sqrt(p*(1.-p))*(3.+2.*p-8.*p**2)
                if p > 0.5:
                    lambdad[ndxuse[edge]] -= 2./3.
                notused6 = np.where(z[ndxuse] != 1.-p)
                if np.size(notused6) == 0:
                    return finish(p,z,u1,u2,lambdae,lambdad,etad)
                ndxuse = ndxuse[notused6[0]]

            ## Case 10 - origin of planet hits origin of star
            origin = np.where(z[ndxuse] == 0)#, complement=notused7)
            if np.size(origin) != 0:
                ## lambda_6
                lambdad[ndxuse[origin]] = -2./3.*(1.-p**2)**1.5
                notused7 = np.where(z[ndxuse] != 0)
                if np.size(notused7) == 0:
                    return finish(p,z,u1,u2,lambdae,lambdad,etad)
                ndxuse = ndxuse[notused7[0]]
   
            q=np.sqrt((x2[ndxuse]-x1[ndxuse])/(1.-x1[ndxuse]))
            n=x2[ndxuse]/x1[ndxuse]-1.
            Ek,Kk = ellke(q)    

            ## Case 3, Case 9 - anywhere in between
            ## lambda_2
            lambdad[ndxuse] = 2./9./np.pi/np.sqrt(1.-x1[ndxuse])*\
                              ((1.-5.*z[ndxuse]**2+p**2+x3[ndxuse]**2)*Kk+\
                               (1.-x1[ndxuse])*(z[ndxuse]**2+7.*p**2-4.)*Ek-\
                               3.*x3[ndxuse]/x1[ndxuse]*ellpic_bulirsch(n,q))

        ## if there are still unused elements, there's a bug in the code
        ## (please report it)
        #notused5 = where(z[notusedyet] > (1.-p))
        if notused5[0] != 0:
            logging.error("The following values of z didn't fit into a case:")

        return finish(p,z,u1,u2,lambdae,lambdad,etad)

# Computes Hasting's polynomial approximation for the complete
# elliptic integral of the first (ek) and second (kk) kind
def ellke(k):
    m1=1.-k**2
    logm1 = np.log(m1)

    a1=0.44325141463
    a2=0.06260601220
    a3=0.04757383546
    a4=0.01736506451
    b1=0.24998368310
    b2=0.09200180037
    b3=0.04069697526
    b4=0.00526449639
    ee1=1.+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
    ee2=m1*(b1+m1*(b2+m1*(b3+m1*b4)))*(-logm1)
    ek = ee1+ee2
        
    a0=1.38629436112
    a1=0.09666344259
    a2=0.03590092383
    a3=0.03742563713
    a4=0.01451196212
    b0=0.5
    b1=0.12498593597
    b2=0.06880248576
    b3=0.03328355346
    b4=0.00441787012
    ek1=a0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
    ek2=(b0+m1*(b1+m1*(b2+m1*(b3+m1*b4))))*logm1
    kk = ek1-ek2
    
    return [ek,kk]

# Computes the complete elliptical integral of the third kind using
# the algorithm of Bulirsch (1965):
def ellpic_bulirsch(n,k):
    kc=np.sqrt(1.-k**2); p=n+1.
    if(p.min() < 0.):
        logging.warning('Negative p')
    m0=1.; c=1.; p=np.sqrt(p); d=1./p; e=kc
    while 1:
        f = c; c = d/p+c; g = e/p; d = 2.*(f*g+d)
        p = g + p; g = m0; m0 = kc + m0
        if (np.absolute(1.-kc/g)).max() > 1.e-8:
            kc = 2*np.sqrt(e); e=kc*m0
        else:
            return 0.5*np.pi*(c*m0+d)/(m0*(m0+p))


#def traptransit(ts,p):
#    return traptransit(ts,p)

def fit_traptransit(ts,fs,p0):
    """
    Fits trapezoid model to provided ts,fs
    """
    pfit,success = leastsq(traptransit_resid,p0,args=(ts,fs))
    if success not in [1,2,3,4]:
        raise NoFitError
    #logging.debug('success = {}'.format(success))
    return pfit

class TraptransitModel(object):
    """
    Model to enable MCMC fitting of trapezoidal shape.
    """
    def __init__(self,ts,fs,sigs=1e-4,maxslope=30):
        self.n = np.size(ts)
        if np.size(sigs)==1:
            sigs = np.ones(self.n)*sigs
        self.ts = ts
        self.fs = fs
        self.sigs = sigs
        self.maxslope = maxslope
        
    def __call__(self,pars):
        pars = np.array(pars)
        return traptransit_lhood(pars,self.ts,self.fs,self.sigs,maxslope=self.maxslope)

def traptransit_lhood(pars,ts,fs,sigs,maxslope=30):
    if pars[0] < 0 or pars[1] < 0 or pars[2] < 2 or pars[2] > maxslope:
        return -np.inf
    resid = traptransit_resid(pars,ts,fs)
    return (-0.5*resid**2/sigs**2).sum()

def traptransit_MCMC(ts,fs,dfs=1e-5,nwalkers=200,nburn=300,niter=1000,
                     threads=1,p0=[0.1,0.1,3,0],return_sampler=False,
                     maxslope=30):
    """
    Fit trapezoidal model to provided ts, fs, [dfs] using MCMC.

    Standard emcee usage.
    """
    model = TraptransitModel(ts,fs,dfs,maxslope=maxslope)
    sampler = emcee.EnsembleSampler(nwalkers,4,model,threads=threads)
    T0 = p0[0]*(1+rand.normal(size=nwalkers)*0.1)
    d0 = p0[1]*(1+rand.normal(size=nwalkers)*0.1)
    slope0 = p0[2]*(1+rand.normal(size=nwalkers)*0.1)
    ep0 = p0[3]+rand.normal(size=nwalkers)*0.0001

    p0 = np.array([T0,d0,slope0,ep0]).T

    pos, prob, state = sampler.run_mcmc(p0, nburn)
    sampler.reset()
    sampler.run_mcmc(pos, niter, rstate0=state)
    if return_sampler:
        return sampler
    else:
        return sampler.flatchain[:,0],sampler.flatchain[:,1],sampler.flatchain[:,2],sampler.flatchain[:,3]

##### Custom Exceptions

class NoEclipseError(Exception):
    pass

class NoFitError(Exception):
    pass

class EmptyPopulationError(Exception):
    pass

class NotImplementedError(Exception):
    pass

class AllWithinRocheError(Exception):
    pass




try:
    import matplotlib.pyplot as plt
    import numpy as np
    import logging
except ImportError:
    plt = None
    np = None
    
    
def setfig(fig=None,**kwargs):
    """
    Sets figure to 'fig' and clears; if fig is 0, does nothing (e.g. for overplotting)

    if fig is None (or anything else), creates new figure
    
    I use this for basically every function I write to make a plot.
    I give the function
    a "fig=None" kw argument, so that it will by default create a new figure.

    .. note::

      There's most certainly a better, more object-oriented
      way of going about writing functions that make figures, but
      this was put together before I knew how to think that way,
      so this stays for now as a convenience.
      
    """
    if fig:
        plt.figure(fig,**kwargs)
        plt.clf()
    elif fig==0:
        pass
    else:
        plt.figure(**kwargs)

def plot2dhist(xdata,ydata,cmap='binary',interpolation='nearest',
               fig=None,logscale=True,xbins=None,ybins=None,
               nbins=50,pts_only=False,**kwargs):
    """Plots a 2d density histogram of provided data

    :param xdata,ydata: (array-like)
        Data to plot.

    :param cmap: (optional)
        Colormap to use for density plot.

    :param interpolation: (optional)
        Interpolation scheme for display (passed to ``plt.imshow``).

    :param fig: (optional)
        Argument passed to :func:`setfig`.

    :param logscale: (optional)
        If ``True`` then the colormap will be based on a logarithmic
        scale, rather than linear.

    :param xbins,ybins: (optional)
        Bin edges to use (if ``None``, then use ``np.histogram2d`` to
        find bins automatically).

    :param nbins: (optional)
        Number of bins to use (if ``None``, then use ``np.histogram2d`` to
        find bins automatically).

    :param pts_only: (optional)
        If ``True``, then just a scatter plot of the points is made,
        rather than the density plot.

    :param **kwargs:
        Keyword arguments passed either to ``plt.plot`` or ``plt.imshow``
        depending upon whether ``pts_only`` is set to ``True`` or not.
        
    """

    setfig(fig)
    if pts_only:
        plt.plot(xdata,ydata,**kwargs)
        return

    ok = (~np.isnan(xdata) & ~np.isnan(ydata) & 
           ~np.isinf(xdata) & ~np.isinf(ydata))
    if ~ok.sum() > 0:
        logging.warning('{} x values and {} y values are nan'.format(np.isnan(xdata).sum(),
                                                                     np.isnan(ydata).sum()))
        logging.warning('{} x values and {} y values are inf'.format(np.isinf(xdata).sum(),
                                                                     np.isinf(ydata).sum()))

    if xbins is not None and ybins is not None:
        H,xs,ys = np.histogram2d(xdata[ok],ydata[ok],bins=(xbins,ybins))
    else:
        H,xs,ys = np.histogram2d(xdata[ok],ydata[ok],bins=nbins)        
    H = H.T

    if logscale:
        H = np.log(H)

    extent = [xs[0],xs[-1],ys[0],ys[-1]]
    plt.imshow(H,extent=extent,interpolation=interpolation,
               aspect='auto',cmap=cmap,origin='lower',**kwargs)
    



from __future__ import print_function, division

import os, os.path
from vespa.fpp import FPPCalculation

import pkg_resources

FOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__)))

def test_fpp(folder=FOLDER):
    f = FPPCalculation.from_ini(folder, n=100)
    f.FPP()





try:
    import numpy as np
    import hashlib

    from hashlib import sha1

    from numpy import all, array, uint8
except ImportError:
    np, hashlib, sha1 = (None, None, None)
    all, array, uint8 = (None, None, None)
    
class hashable(object):
    r'''Hashable wrapper for ndarray objects.

        Instances of ndarray are not hashable, meaning they cannot be added to
        sets, nor used as keys in dictionaries. This is by design - ndarray
        objects are mutable, and therefore cannot reliably implement the
        __hash__() method.

        The hashable class allows a way around this limitation. It implements
        the required methods for hashable objects in terms of an encapsulated
        ndarray object. This can be either a copied instance (which is safer)
        or the original object (which requires the user to be careful enough
        not to modify it).

        This class taken from `here <http://stackoverflow.com/questions/1939228/constructing-a-python-set-from-a-numpy-matrix/5173201#5173201>`_; edited only slightly.
    '''
    def __init__(self, wrapped, tight=False):
        r'''Creates a new hashable object encapsulating an ndarray.

            wrapped
                The wrapped ndarray.

            tight
                Optional. If True, a copy of the input ndaray is created.
                Defaults to False.
        '''
        self.__tight = tight
        self.__wrapped = array(wrapped) if tight else wrapped
        #self.__hash = int(sha1(wrapped.view(uint8)).hexdigest(), 16)
        self.__hash = int(sha1(np.ascontiguousarray(wrapped)).hexdigest(), 16)

    def __eq__(self, other):
        return all(self.__wrapped == other.__wrapped)

    def __hash__(self):
        return self.__hash

    def unwrap(self):
        r'''Returns the encapsulated ndarray.

            If the wrapper is "tight", a copy of the encapsulated ndarray is
            returned. Otherwise, the encapsulated ndarray itself is returned.
        '''
        if self.__tight:
            return array(self.__wrapped)

        return self.__wrapped

def hasharray(arr):
    """
    Hashes array-like object (except DataFrame)
    """
    #return hash(hashlib.sha1(np.ascontiguousarray(arr)).hexdigest())
    return hash(hashable(np.array(arr)))

def hashdf(df):
    """hashes a pandas dataframe, forcing values to float
    """
    return hasharray(df.values.astype(float))

def hashcombine(*xs):
    """
    Combines multiple hashes using xor
    """
    k = 0
    for x in xs:
        k ^= hash(x)
    k ^= hash(xs)
    return k

def hashdict(d):
    """Hash a dictionary
    """
    k = 0
    for key,val in d.iteritems():
        k ^= hash(key) ^ hash(val)
    return k



import numpy as np
cimport numpy as np
cimport cython
from cpython cimport bool

FOO = "Hi" #for weird __init__ bug purposes

DTYPE = np.float
ctypedef np.float_t DTYPE_t
cdef DTYPE_t pi = 3.1415926535897932384626433832795

cdef extern from "math.h":
    double sin(double)
    double cos(double)
    double tanh(double)
    double sqrt(double)
    double atan2(double,double)
    double acos(double)
    double abs(double)
    double log(double)
    double ceil(double)
    
@cython.boundscheck(False)
def bindata(np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] fs, DTYPE_t dt):
    """Requires ts to be ordered
    """
    cdef long npts = len(ts)
    cdef DTYPE_t trange = ts[npts-1]-ts[0]
    cdef long nbins = int(ceil(trange/dt))

    cdef np.ndarray[DTYPE_t] tbin = np.empty(nbins,dtype=float)
    cdef np.ndarray[DTYPE_t] fbin = np.empty(nbins,dtype=float)
    cdef np.ndarray[DTYPE_t] ubin = np.empty(nbins,dtype=float)

    cdef unsigned int i
    cdef unsigned int j = 0
    cdef DTYPE_t lo = ts[0]
    cdef DTYPE_t hi = lo+dt
    cdef DTYPE_t fsum = 0
    cdef DTYPE_t fsumsq = 0
    cdef DTYPE_t tsum = 0
    cdef long count = 0
    for i in range(npts):
        if ts[i] > hi and count > 0:
            tbin[j] = tsum/count
            fbin[j] = fsum/count
            ubin[j] = sqrt(fsumsq/count - (fsum/count)*(fsum/count)) #std in bin
            lo = hi
            hi = lo+dt
            count = 0
            fsum=0
            fsumsq=0
            tsum=0
            j += 1
        else:
            tsum += ts[i]
            fsum += fs[i]
            fsumsq += fs[i]*fs[i]
            count += 1
        if i==npts-1:
            if count==0:
                tbin[j] = tbin[j-1]
                fbin[j] = fbin[j-1]
                ubin[j] = ubin[j-1]
            else:
                tbin[j] = tsum/count
                fbin[j] = fsum/count
                ubin[j] = sqrt(fsumsq/count - (fsum/count)*(fsum/count)) #std in bin
                
    return tbin,fbin,ubin
            

@cython.boundscheck(False)
def find_eclipse(np.ndarray[DTYPE_t] Es, DTYPE_t b,
                 DTYPE_t aR, DTYPE_t ecc, DTYPE_t w, DTYPE_t width, bool sec):
    cdef long npts = len(Es)
    w = w*pi/180. # degrees to radians
    cdef DTYPE_t inc
    if sec:
        inc = acos(b/aR*(1-ecc*sin(w))/(1-ecc*ecc))
    else:
        inc = acos(b/aR*(1+ecc*sin(w))/(1-ecc*ecc))
    cdef DTYPE_t nu,r,X,Y
    cdef DTYPE_t xmin = 10000.
    cdef bool on_rightside
    cdef np.ndarray[DTYPE_t] zs = np.empty(npts,dtype=float)
    cdef np.ndarray[long] in_eclipse = np.empty(npts,dtype=np.int) #how to make boolean array?
    cdef unsigned int i
    for i in range(npts):
        nu = 2 * atan2(sqrt(1+ecc)*sin(Es[i]/2),sqrt(1-ecc)*cos(Es[i]/2)) #true anomaly
        r = aR*(1-ecc*ecc)/(1+ecc*cos(nu))  #secondary distance from primary in units of R1
        X = -r*cos(w + nu)    
        Y = -r*sin(w + nu)*cos(inc)
        zs[i] = sqrt(X*X + Y*Y)
        if sec:
            on_rightside = (sin(nu + w) < 0)
        else:
            on_rightside = (sin(nu + w) >= 0)
        in_eclipse[i] = (on_rightside)*(zs[i] < width)
    return zs,in_eclipse
        
@cython.boundscheck(False)
def traptransit(np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] pars):
    """pars = [T,delta,T_over_tau,tc]"""
    cdef DTYPE_t t1 = pars[3] - pars[0]/2.
    cdef DTYPE_t t2 = pars[3] - pars[0]/2. + pars[0]/pars[2]
    cdef DTYPE_t t3 = pars[3] + pars[0]/2. - pars[0]/pars[2]
    cdef DTYPE_t t4 = pars[3] + pars[0]/2.
    cdef long npts = len(ts)
    cdef np.ndarray[DTYPE_t] fs = np.empty(npts,dtype=float)
    cdef unsigned int i
    for i in range(npts):
        if (ts[i]  > t1) and (ts[i] < t2):
            fs[i] = 1-pars[1]*pars[2]/pars[0]*(ts[i] - t1)
        elif (ts[i] > t2) and (ts[i] < t3):
            fs[i] = 1-pars[1]
        elif (ts[i] > t3) and (ts[i] < t4):
            fs[i] = 1-pars[1] + pars[1]*pars[2]/pars[0]*(ts[i]-t3)
        else:
            fs[i] = 1
    return fs

@cython.boundscheck(False)
def traptransit_resid( np.ndarray[DTYPE_t] pars, np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] fs):
    cdef long npts = len(ts)
    cdef np.ndarray[DTYPE_t] resid = np.empty(npts,dtype=float)
    cdef np.ndarray[DTYPE_t] fmod = traptransit(ts,pars)
    cdef unsigned int i
    for i in range(npts):
        resid[i] = fmod[i]-fs[i]
    return resid

@cython.boundscheck(False)
def traptransit_lhood(np.ndarray[DTYPE_t] pars, np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] fs, 
                      np.ndarray[DTYPE_t] sigs):
    cdef long npts = len(ts)
    cdef DTYPE_t lhood = 0
    cdef np.ndarray[DTYPE_t] fmod = traptransit(ts,pars)
    cdef unsigned int i
    for i in range(npts):
        lhood += -0.5*(fs[i]-fmod[i])*(fs[i]-fmod[i])/(sigs[i]*sigs[i])
    return lhood
    


@cython.boundscheck(False)
def protopapas(np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] pars):
    """pars: eta,theta,c,tau"""
    cdef DTYPE_t T = 10000.
    cdef long npts = len(ts)
    cdef np.ndarray[DTYPE_t] fs = np.empty(npts,dtype=float)
    cdef DTYPE_t tp = 0.
    cdef unsigned int i
    for i in range(npts):
        tp = T*sin(pi*(ts[i]-pars[3])/T)/pi/pars[0]
        fs[i] = 1 - pars[1] + 0.5*pars[1]*(2 - tanh(pars[2]*(tp+0.5)) + tanh(pars[2]*(tp-0.5)))
    return fs

@cython.boundscheck(False)
def protopapas_resid(np.ndarray[DTYPE_t] pars, np.ndarray[DTYPE_t] ts, np.ndarray[DTYPE_t] fs):
    cdef long npts = len(ts)
    cdef np.ndarray[DTYPE_t] resid = np.empty(npts,dtype=float)
    cdef np.ndarray[DTYPE_t] fmod = protopapas(ts,pars)
    cdef unsigned int i
    for i in range(npts):
        resid[i] = fmod[i]-fs[i]
    return resid
     

#####Mandel/Agol utility functions below.  UNFINISHED; will not work.

## Computes Hasting's polynomial approximation for the complete
## elliptic integral of the first (ek) and second (kk) kind
#@cython.boundscheck(False)
#def ellke(DTYPE_t k):
#    DTYPE_t m1=1.-k**2
#    DTYPE_t logm1 = log(m1)

#    DTYPE_t a1=0.44325141463
#    DTYPE_t a2=0.06260601220
#    DTYPE_t a3=0.04757383546
#    DTYPE_t a4=0.01736506451
#    DTYPE_t b1=0.24998368310
#    DTYPE_t b2=0.09200180037
#    DTYPE_t b3=0.04069697526
#    DTYPE_t b4=0.00526449639
#    DTYPE_t ee1=1.+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
#    DTYPE_t ee2=m1*(b1+m1*(b2+m1*(b3+m1*b4)))*(-logm1)
#    DTYPE_t ek = ee1+ee2
        
#    DTYPE_t a0=1.38629436112
#    DTYPE_t a1=0.09666344259
#    DTYPE_t a2=0.03590092383
#    DTYPE_t a3=0.03742563713
#    DTYPE_t a4=0.01451196212
#    DTYPE_t b0=0.5
#    DTYPE_t b1=0.12498593597
#    DTYPE_t b2=0.06880248576
#    DTYPE_t b3=0.03328355346
#    DTYPE_t b4=0.00441787012
#    DTYPE_t ek1=a0+m1*(a1+m1*(a2+m1*(a3+m1*a4)))
#    DTYPE_t ek2=(b0+m1*(b1+m1*(b2+m1*(b3+m1*b4))))*logm1
#    DTYPE_t kk = ek1-ek2
    
#    return [ek,kk]

## Computes the complete elliptical integral of the third kind using
## the algorithm of Bulirsch (1965):
#@cython.boundscheck(False)
#def ellpic_bulirsch(DTYPE_t n,DTYPE_t k):
#    DTYPE_t kc=sqrt(1.-k*k)
#    DTYPLE_t p=n+1.
#    if(p.min() < 0.):
#        print 'Negative p'
#    m0=1.; c=1.; p=sqrt(p); d=1./p; e=kc
#    while 1:
#        f = c; c = d/p+c; g = e/p; d = 2.*(f*g+d)
#        p = g + p; g = m0; m0 = kc + m0
#        if (absolute(1.-kc/g)).max() > 1.e-8:
#            kc = 2*sqrt(e); e=kc*m0
#        else:
#            return 0.5*pi*(c*m0+d)/(m0*(m0+p))



from __future__ import print_function,division

import os,os.path
import pkg_resources
import logging

on_rtd = False

try:
    import numpy as np
    import numpy.random as rand
except ImportError:
    on_rtd = True
    np, rand = (None, None)

try:
    from simpledist.distributions import KDE_Distribution
except ImportError:
    logging.warning('simpledist not available.')
    KDE_Distribution = None

if not on_rtd:
    from astropy.units import Quantity
else:
    Quantity = None 

if not on_rtd:
    DATAFOLDER = os.path.abspath(os.path.join(os.path.dirname(__file__), 'data'))

    RAGHAVAN_PERS = np.recfromtxt('{}/raghavan_periods.dat'.format(DATAFOLDER))
    RAGHAVAN_LOGPERS = np.log10(RAGHAVAN_PERS.f1[RAGHAVAN_PERS.f0 == 'Y'])
    RAGHAVAN_BINPERS = RAGHAVAN_PERS.f1[RAGHAVAN_PERS.f0 == 'Y']
    RAGHAVAN_BINPERKDE = KDE_Distribution(RAGHAVAN_BINPERS,adaptive=False)
    RAGHAVAN_LOGPERKDE = KDE_Distribution(RAGHAVAN_LOGPERS,adaptive=False)

    #from Multiple Star Catalog
    MSC_TRIPDATA = np.recfromtxt('{}/multiple_pecc.txt'.format(DATAFOLDER),names=True)
    MSC_TRIPLEPERS = MSC_TRIPDATA.P
    MSC_TRIPPERKDE = KDE_Distribution(MSC_TRIPLEPERS,adaptive=False)
    MSC_TRIPLOGPERKDE = KDE_Distribution(np.log10(MSC_TRIPLEPERS),adaptive=False)
else:
    DATAFOLDER = None
    RAGHAVAN_PERS, RAGHAVAN_LOGPERS, RAGHAVAN_BINPERS = (None, None, None)
    RAGHAVAN_BINPERKDE, RAGHAVAN_LOGPERKDE = (None, None)
    MSC_TRIPDATA, MSC_TRIPLEPERS, MSC_TRIPPERKDE, MSC_TRIPLOGPERKDE = (None, None,
                                                                       None, None)
    
    
    
def randpos_in_circle(n,rad,return_rad=False):
    x = rand.random(n)*2*rad - rad
    y = rand.random(n)*2*rad - rad
    mask = (x**2 + y**2 > rad**2)
    nw = mask.sum()
    while nw > 0:
        x[mask] = rand.random(nw)*2*rad-rad
        y[mask] = rand.random(nw)*2*rad-rad
        mask = (x**2 + y**2 > rad**2)
        nw = mask.sum()
    if return_rad:
        return np.sqrt(x**2 + y**2)
    else:
        return x,y

def draw_pers_eccs(n,**kwargs):
    """
    Draw random periods and eccentricities according to empirical survey data.
    """
    pers = draw_raghavan_periods(n)
    eccs = draw_eccs(n,pers,**kwargs)
    return pers,eccs

def flat_massratio(n, qmin=0.1, qmax=1.):
    return rand.uniform(size=n)*(qmax - qmin) + qmin

def flat_massratio_fn(qmin=0.1,qmax=1.):
    def fn(n):
        return rand.uniform(size=n)*(qmax - qmin) + qmin
    return fn

def draw_raghavan_periods(n):
    """
    Draw orbital periods according to Raghavan (2010)
    """
    logps = RAGHAVAN_LOGPERKDE.resample(n)
    return 10**logps

def draw_msc_periods(n):
    """
    Draw orbital periods according to Multiple Star Catalog
    """
    logps = MSC_TRIPLOGPERKDE.resample(n)
    return 10**logps

def draw_eccs(n,per=10,binsize=0.1,fuzz=0.05,maxecc=0.97):
    """draws eccentricities appropriate to given periods, generated according to empirical data from Multiple Star Catalog
    """
    if np.size(per) == 1 or np.std(np.atleast_1d(per))==0:
        if np.size(per)>1:
            per = per[0]
        if per==0:
            es = np.zeros(n)
        else:
            ne=0
            while ne<10:
                mask = np.absolute(np.log10(MSC_TRIPLEPERS)-np.log10(per))<binsize/2.
                es = MSC_TRIPDATA.e[mask]
                ne = len(es)
                if ne<10:
                    binsize*=1.1
            inds = rand.randint(ne,size=n)
            es = es[inds] * (1 + rand.normal(size=n)*fuzz)    
    else:
        longmask = (per > 25)
        shortmask = (per <= 25)
        es = np.zeros(np.size(per))

        elongs = MSC_TRIPDATA.e[MSC_TRIPLEPERS > 25]
        eshorts = MSC_TRIPDATA.e[MSC_TRIPLEPERS <= 25]

        n = np.size(per)
        nlong = longmask.sum()
        nshort = shortmask.sum()
        nelongs = np.size(elongs)
        neshorts = np.size(eshorts)
        ilongs = rand.randint(nelongs,size=nlong)
        ishorts = rand.randint(neshorts,size=nshort)
        
        es[longmask] = elongs[ilongs]
        es[shortmask] = eshorts[ishorts]

    es = es * (1 + rand.normal(size=n)*fuzz)
    es[es>maxecc] = maxecc
    return np.absolute(es)

########## other utility functions; copied from old code

if not on_rtd:
    import astropy.constants as const
    AU = const.au.cgs.value
    RSUN = const.R_sun.cgs.value
    MSUN = const.M_sun.cgs.value
    DAY = 86400 #seconds
    G = const.G.cgs.value
else:
    const, AU, RSUN, MSUN, DAY, G = (None, None, None, None, None, None)
    
def rochelobe(q):
    """returns r1/a; q = M1/M2"""
    return 0.49*q**(2./3)/(0.6*q**(2./3) + np.log(1+q**(1./3)))

def withinroche(semimajors,M1,R1,M2,R2):
    """
    Returns boolean array that is True where two stars are within Roche lobe
    """
    q = M1/M2
    return ((R1+R2)*RSUN) > (rochelobe(q)*semimajors*AU)
    
def semimajor(P,mstar=1):
    """Returns semimajor axis given P in days, mstar in solar masses.
    """
    return ((P*DAY/2/np.pi)**2*G*mstar*MSUN)**(1./3)/AU

def period_from_a(a,mstar):
    return np.sqrt(4*np.pi**2*(a*AU)**3/(G*mstar*MSUN))/DAY

def addmags(*mags):
    """
    "Adds" magnitudes.  Yay astronomical units!
    """
    tot=0
    for mag in mags:
        tot += 10**(-0.4*mag)
    return -2.5*np.log10(tot)

def fluxfrac(*mags):
    """Returns fraction of total flux in first argument, assuming all are magnitudes.
    """
    Ftot = 0
    for mag in mags:
        Ftot += 10**(-0.4*mag)
    F1 = 10**(-0.4*mags[0])
    return F1/Ftot

def dfromdm(dm):
    """Returns distance given distance modulus.
    """
    if np.size(dm)>1:
        dm = np.atleast_1d(dm)
    return 10**(1+dm/5)

def distancemodulus(d):
    """Returns distance modulus given d in parsec.
    """
    if type(d)==Quantity:
        x = d.to('pc').value
    else:
        x = d #assumed to be pc

    if np.size(x)>1:
        d = np.atleast_1d(x)
    return 5*np.log10(x/10)

def fbofm(M):
    return 0.45 - (0.7-M)/4


def mult_masses(mA, f_binary=0.4, f_triple=0.12,
                minmass=0.11, qmin=0.1, n=1e5):
    """Returns m1, m2, and m3 appropriate for TripleStarPopulation, given "primary" mass (most massive of system) and binary/triple fractions.
                

    star with m1 orbits (m2 + m3).  This means that the primary mass mA will correspond
    either to m1 or m2.  Any mass set to 0 means that component does not exist.
    """
    
    if np.size(mA) > 1:
        n = len(mA)
    else:
        mA = np.ones(n) * mA


    r = rand.random(n)
    is_single = r > (f_binary + f_triple)
    is_double = (r > f_triple) & (r < (f_binary + f_triple))
    is_triple = r <= f_triple

    CwA = rand.random(n) < 0.5
    CwB = ~CwA
        

    #these for Triples:
    
    minq2_A = minmass/mA
    q2_A = rand.random(n)*(1-minq2_A) + minq2_A
    minq1_A = (minmass/mA)/(1+q2_A)
    maxq1_A = 1/(1+q2_A)
    q1_A = rand.random(n)*(maxq1_A-minq1_A) + minq1_A

    minq1_B = 2*minmass/mA
    q1_B = rand.random(n)*(1-minq1_B) + minq1_B
    minq2_B = np.maximum(((q1_B*mA)-minmass)/minmass,
                         (q1_B*mA - minmass)/(q1_B*mA + minmass))
    maxq2_B = 1.
    q2_B = rand.random(n)*(maxq2_B-minq2_B) + minq2_B

    
    mB_A = q1_A*(1 + q2_A) * mA
    mC_A = q2_A * mA

    mB_B = (q1_B/(1 + q2_B)) * mA
    mC_B = (q1_B*q2_B)/(1 + q2_B) * mA

    mB = CwA*mB_A + CwB*mB_B
    mC = CwA*mC_A + CwB*mC_B

    #for binaries-only
    qmin = minmass/mA
    q = rand.random(n)*(1-qmin) + qmin    
    mB[is_double] = q[is_double]*mA[is_double]

                
    #now need to define the proper mapping from A,B,C to 1,2,3:
    # If no B or C present, then A=1
    # If B present but not C, then A=1, B=2
    # If both B and C present then:
    #     If C is with A, then A=2, C=3, B=1
    #     If C is with B, then A=1, B=2, C=3
    m1 = (mA)*(is_single + is_double) + (CwA*mB + CwB*mA)*is_triple
    m2 = (mB)*is_double + (CwA*mA + CwB*mB)*is_triple
    m3 = mC*is_triple
    
    return m1, m2, m3
    



#! /usr/bin/env python
from __future__ import print_function, division

import logging
import sys, re, os

try:
    import numpy as np
    import pandas as pd
except ImportError:
    np, pd = (None, None)
    
from .transit_basic import eclipse_pars, eclipse_tt
from .transit_basic import NoEclipseError, NoFitError
try:
    from progressbar import Percentage,Bar,RotatingMarker,ETA,ProgressBar
    pbar_ok = True
except ImportError:
    logging.warning('progressbar not imported')
    pbar_ok = False
    
def fitebs(data, MAfn=None, conv=True, cadence=0.020434028,
           use_pbar=True, msg=''): 
    """Fits trapezoidal shape to eclipses described by parameters in data

    Takes a few minutes for 20,000 simulations. 
           
    :param data:
        Input ``DataFrame`` holding data.  Must have following columns:
        ``'P', 'mass_1', 'mass_2', 'radius_1', 'radius_2'``, 
        ``'inc', 'ecc', 'w', 'dpri', 'dsec', 'b_sec', 'b_pri'``,
        ``'fluxfrac_1', 'fluxfrac_2', 'u1_1', 'u1_2', 'u2_1', 'u2_2'``.

    :param MAfn:
        :class:`MAInterpolationFunction` object; if not passed, it will be
        created.

    :param conv:
        Whether to convolve theoretical transit shape
        with box-car filter to simulate observation integration time.

    :param cadence:
        Integration time used if ``conv`` is ``True``.  Defaults to
        Kepler mission cadence.

    :param use_pbar:
        Whether to use nifty visual progressbar when doing calculation.

    :param msg:
        Message to print for pbar.

    :return df:
        Returns dataframe with the following columns:
        ``'depth', 'duration', 'slope'``,
        ``'secdepth', 'secondary'``.

    """
    n = len(data)


    p0s, bs, aRs = eclipse_pars(data['P'], data['mass_1'], data['mass_2'],
                                data['radius_1'], data['radius_2'], 
                                inc=data['inc'],
                                ecc=data['ecc'], w=data['w'])

    deps, durs, slopes = (np.zeros(n), np.zeros(n), np.zeros(n))
    secs = np.zeros(n).astype(bool)
    dsec = np.zeros(n)

    if use_pbar and pbar_ok:
        widgets = [msg+'fitting shape parameters for %i systems: ' % n,Percentage(),
                   ' ',Bar(marker=RotatingMarker()),' ',ETA()]
        pbar = ProgressBar(widgets=widgets,maxval=n)
        pbar.start()

    for i in xrange(n):
        pri = (data['dpri'][i] > data['dsec'][i]) or np.isnan(data['dsec'][i])
        sec = not pri
        secs[i] = sec
        p0, aR = (p0s[i], aRs[i])
        if sec:
            b = data['b_sec'][i]
            frac = data['fluxfrac_2'][i]
            dsec[i] = data['dpri'][i]
            u1 = data['u1_2'][i]
            u2 = data['u2_2'][i]
        else:
            b = data['b_pri'][i]
            frac = data['fluxfrac_1'][i]
            dsec[i] = data['dsec'][i]
            u1 = data['u1_1'][i]
            u2 = data['u2_1'][i]
        try:
            if MAfn is not None:
                if p0 > MAfn.pmax or p0 < MAfn.pmin:
                    trap_pars = eclipse_tt(p0,b,aR,data['P'][i],conv=conv,MAfn=None,
                                              cadence=cadence, frac=frac,
                                              ecc=data['ecc'][i],w=data['w'][i],
                                              sec=sec,u1=u1,u2=u2)
                else:
                    trap_pars = eclipse_tt(p0,b,aR,data['P'][i],conv=conv,MAfn=MAfn,
                                              cadence=cadence, frac=frac,
                                              ecc=data['ecc'][i],w=data['w'][i],
                                              sec=sec,u1=u1,u2=u2)
            else:
                trap_pars = eclipse_tt(p0,b,aR,data['P'][i],conv=conv,MAfn=MAfn,
                                          cadence=cadence, frac=frac,
                                          ecc=data['ecc'][i],w=data['w'][i],
                                          sec=sec,u1=u1,u2=u2)
            #logging.debug('{}'.format(trap_pars))
            
            durs[i], deps[i], slopes[i] = trap_pars
            if use_pbar and pbar_ok:
                pbar.update(i)

        except NoEclipseError:
            logging.error('No eclipse registered for index {}'.format(i))
            continue
        except NoFitError:
            logging.error('Fit did not converge for index {}'.format(i))
            continue
        except KeyboardInterrupt:
            raise
        except:
            logging.error('unknown error for index {}'.format(i))
            raise
            continue

    return pd.DataFrame({'depth':deps, 'duration':durs,
                         'slope':slopes, 'secdepth':dsec,
                         'secondary':secs})



from .populations import StarPopulation
from .populations import MultipleStarPopulation, ColormatchMultipleStarPopulation
from .populations import Spectroscopic_MultipleStarPopulation
from .populations import BGStarPopulation, BGStarPopulation_TRILEGAL
from .populations import BinaryPopulation, Simulated_BinaryPopulation
from .populations import Raghavan_BinaryPopulation, TriplePopulation
from .populations import Observed_TriplePopulation, Observed_BinaryPopulation

from .extinction import get_AV_infinity
from .trilegal import get_trilegal




from setuptools import setup, Extension, find_packages

on_rtd = False
try:
    from Cython.Distutils import build_ext
    import numpy
except ImportError:
    on_rtd = True
    numpy = None
    build_ext = None
    
import os

def readme():
    with open('README.rst') as f:
        return f.read()

# Hackishly inject a constant into builtins to enable importing of the
# package before the library is built.
import sys
if sys.version_info[0] < 3:
    import __builtin__ as builtins
else:
    import builtins
builtins.__VESPA_SETUP__ = True
import vespa
version = vespa.__version__


# Publish the library to PyPI.
if "publish" in sys.argv[-1]:
    os.system("python setup.py sdist upload")
    sys.exit()

# Push a new tag to GitHub.
if "tag" in sys.argv:
    os.system("git tag -a {0} -m 'version {0}'".format(version))
    os.system("git push --tags")
    sys.exit()

if not on_rtd:
    transit_utils = [Extension('vespa_transitutils',['vespa/vespa_transitutils.pyx'],
                                include_dirs=[numpy.get_include()])]
else:
    transit_utils = None
        
setup(name = "VESPA",
      version = version,
      description = "Calculate astrophysical false positive probabilities for transiting exoplanet signals",
      long_description = readme(),
      author = "Timothy D. Morton",
      author_email = "tim.morton@gmail.com",
      url = "https://github.com/timothydmorton/VESPA",
      #packages = ['vespa', 'vespa/stars',
      #            'vespa/orbits'],
      packages = find_packages(),
      package_data = {'vespa': ['data/*', 'tests/*.ini',
                                'tests/*.h5', 'tests/*.pkl'],
                      'vespa.stars': ['data/*'],
                      'vespa.orbits':['data/*']},
      ext_modules = transit_utils,
      scripts = ['scripts/get_trilegal',
                 'scripts/koifpp',
                 'scripts/batch_koifpp_condor',
                 'scripts/calcfpp',
                 'scripts/koifpp-config'],
      cmdclass = {'build_ext': build_ext},
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Astronomy'
        ],
      install_requires=['cython','pandas>=0.13','simpledist>=0.1.11', 'emcee', 
                        'isochrones>=0.8.1', 'acor'],
      zip_safe=False
) 
